#!/usr/bin/env python

from Tkinter import *
from types import *
import re, string, time, os, sys

import pygale.engine
if sys.platform != 'win32':
	pygale.engine.engine = pygale.engine.TkEngine()
from pygale import *
import thumbcache

if sys.version >= '1.6':
	STRING_TYPES = [StringType, UnicodeType]
else:
	STRING_TYPES = [StringType]

# Set by ui.py
Config = None

# URL matching pattern
URLPAT = 'https?:[\w/\.:;+\-~\%#?=&,]+|www\.[\w/\.:;+\-~\%#?=&,]+|' +\
	'ftp:[\w/\.:;+\-~\%#?=&,]+'

# Encapsulate information about a single connection to the Gale server
class Connection:
	def __init__(self, name, sub):
		self._name = name		# Name of tab
		self._dosub = 1		# Whether to subscribe or not
		self._sub = sub		# Subscription
		self._master = None		# Pointer to a GaleWin
		self._galec = None		# Pointer to Gale client
		self._widget = None		# Pointer to Text widget
		self._socketfd = -1		# Socket descriptor
		self._pufflist = []		# List of puffs
		self._curpuff = None	# Currently selected puff (index into
							# pufflist)
		self._savepuffs = 0		# Save the puffs to disk
		self._logsent = 0		# Log sent puffs (0=no, 1=priv, 2=pub+priv)
		self._ccself = 0		# CC self on private puffs
		self._send_receipts = 1	# Send return receipts to private puffs

		# On new puff:
		self._beep = 0			# Beep
		self._command = None	# Run command
		self._deiconify = 0		# Deiconify window
		self._highlight = 0		# Highlight tab
		self._lift = 0			# Lift tab

		# State vars
		self._temp_hash = {}
	
#	# For pickling
#	def __getstate__(self):
#		state = {}
#		saved_attribs = ['name', 'dosub', 'sub', 'beep',
#			'command', 'deiconify', 'highlight', 'lift', 'savepuffs',
#			'logsent']
#		if self._savepuffs:
#			saved_attribs.append('pufflist')
#		for k in saved_attribs:
#			uname = '_' + k
#			state[uname] = self.__dict__[uname]
#		return state
#
#	def __setstate__(self, dict):
#		self.__init__(dict['_name'], dict['_sub'])
#		for k in dict.keys():
#			# Change old flash settings to new highlight settings
#			# Get rid of this when no one has an old config file anymore
#			if k == '_flash':
#				self.__dict__['_highlight'] = dict['_flash']
#			else:
#				self.__dict__[k] = dict[k]

	# Accessors for config variables
	# Reconnect manually if necessary
	def __setitem__(self, key, val):
		uname = '_' + key
		self.__dict__[uname] = val

	# Getitem does double duty: using an integer index returns the item in
	# the pufflist, using a string index returns the config variable
	def __getitem__(self, key):
		if type(key) is IntType:
			return self._pufflist[key]
		else:
			uname = '_' + key
			return self.__dict__[uname]

	# Get our length as if we were the pufflist
	def __len__(self):
		return len(self._pufflist)

	# Accessors for other variables
	def widget(self):
		return self._widget

	def client(self):
		return self._galec
	
	# Store this master and widget into this connection
	def setwidget(self, master, widget):
		self._master = master
		self._widget = widget
		self._origbg = self._widget.cget('text_background')
	
	def disconnect(self):
		self._galec.disconnect()

	# TODO: this is suboptimal when _dosub is set to 0
	# Maybe I should just get rid of _dosub altogether.
	# It makes a connection to the galed in that case, but subscribes to
	# "-".  It really shouldn't connect in that case.
	# Make connection to server
	def connect(self):
		self.make_connection()
		if sys.platform == 'win32':
			self.check_socketread()

		if self._dosub:
			sub = string.strip(re.sub('\n|\r|\t', '', self._sub))
		else:
			sub = None
	
		self._widget.configure(text_state = 'normal')
		if sub is not None:
			self._widget.insert('end', 'Subscribed to ' + sub + '\n')
		else:
			self._widget.insert('end', 'Unsubscribed\n')
		self._widget.see('end')
		self._widget.configure(text_state = 'disabled')

	# Our subscription changed, reconnect
	def reconnect(self):
		if self._dosub:
			sub = string.strip(re.sub('\n|\r|\t', '', self._sub))
		else:
			sub = None
		self._galec.sub_to(sub)

		self._widget.configure(text_state = 'normal')
		if sub is not None:
			self._widget.insert('end', 'Subscribed to ' + sub + '\n')
		else:
			self._widget.insert('end', 'Unsubscribed\n')
		self._widget.see('end')
		self._widget.configure(text_state = 'disabled')
		
	def make_connection(self):
		if self._dosub:
			sub = string.strip(re.sub('\n|\r|\t', '', self._sub))
		else:
			sub = None

		# Set up connection to Gale server
		if self._socketfd != -1:
			self.delete_filehandler()
		self._galec = pygale.GaleClient()
		self._galec.set_onconnect(lambda h, s=self, sub=sub:
			s.make_subs(h, sub))
		self._galec.set_ondisconnect(self.disconnected)
		self._galec.connect(lambda h, s=self: s.make_connection2(h))
	
	def make_connection2(self, host):
		if host is None:
			# Couldn't connect to server
			self._master.set_status('Unable to connect to a gale server')
			return
	
	def make_subs(self, host, sub):
		self._galec.sub_to(sub)
		self._socketfd = self._galec.socket()
		self.register_filehandler()

		self._widget.configure(text_state='normal')
		self._widget.insert('end', 'Connected to server at %s\n' % host)
		self._widget.configure(text_state='disabled')
	
	def disconnected(self):
		self.delete_filehandler()
		self._widget.configure(text_state='normal')
		self._widget.insert('end', 'Disconnected from server\n')
		self._widget.configure(text_state='disabled')
	
	# ------------------------------------------------------------
	# Icky workarounds for Win32
	def register_filehandler(self):
		self._galec.set_puff_callback(lambda p, s=self: s._master.recvpuff(s, p))
		self._galec.set_verify_callback(lambda p, s=self: s._master.verify_update(s, p))
		engine.timeout.set_timer_callback(self.timer_callback)

	def delete_filehandler(self):
		self._galec.del_puff_callback()
	
	def check_socketread(self, *args):
		engine.engine.process(0)
		self._master.after(500, self.check_socketread)

	def timer_callback(self, *args):
		next_timeout = engine.timeout.next_timeout()
		if next_timeout is None:
			return
		time_til_timeout = next_timeout - time.time()
		if (time_til_timeout <= 0):
			engine.timeout.process()
			return
		self._master.after(int(time_til_timeout * 1000), self.timer_callback)

	# ------------------------------------------------------------

	def pufflist(self):
		return self._pufflist

	def set_curpuff(self, index):
		if self._curpuff is not None:
			self._widget.tag_config(
				'pufftag-%i' % hash(self._pufflist[self._curpuff]),
				background=self._origbg)
		puff = self._pufflist[index]
		tag = 'pufftag-%i' % hash(puff)
		self._curpuff = index
		self._widget.tag_config(tag, background=Config.HIGHLIGHT)
		# Fix due to MKF: the tag ranges might be discontiguous
		ranges = self._widget.tag_ranges(tag)
		start = ranges[0]
		end = ranges[-1]
		self._widget.see(end)
		self._widget.see(start)
	
	def set_curpuff_frompuff(self, puff):
		self._master.set_pub_focus()
		index = self._pufflist.index(puff)
		self.set_curpuff(index)
	
	def curpuff(self):
		if self._curpuff is not None:
			return self._pufflist[self._curpuff]
		else:
			return None
	
	def curindex(self):
		return self._curpuff

	def addpuff(self, msg, new=1):
		# Filter out /ping first, before anything else
		cats = string.split(msg.get_cat(), ':')
		for c in cats:
			if c == '/ping':
				return

		self._widget.configure(text_state = 'normal')
		self._temp_hash[msg] = 1
		self._pufflist.append(msg)
		tag = 'pufftag-%i' % hash(msg)
		signer_mark_begin = 'signer-mark-%i-begin' % hash(msg)

		# Indent margins
		self._widget.tag_config(tag, lmargin1=4, lmargin2=4, rmargin=4)

		# The previous line created the latest tag
		# Now raise the sel tag so that selections are always drawn on
		# top of any other coloration
		self._widget.tag_raise('quotation')
		self._widget.tag_raise('sel')

		# Check for this condition before adding all this text below
		if self._widget.dlineinfo('end - 2 lines'):
			scroll_bottom = 1
		else:
			scroll_bottom = 0

		bodytext = msg.get('message/body', '')
		bodytext = re.sub('\r\n', '\n', bodytext)

		# Format return receipts specially
		if msg.get('answer/receipt', None) is not None or\
			msg.get('notice/presence', None) is not None:
			# Return receipts are never "new"
			if msg.get('answer/receipt', None) is not None:
				new = 0
			self._widget.insert('end',
				'* %s ' %
				time.strftime('%Y-%m-%d %H:%M:%S',
				time.localtime(msg.get('id/time', time.time()))),
				tag)
			if msg.get('answer/receipt', None) is not None:
				self._widget.insert('end', 'received: ', tag)
			if msg.get('notice/presence', None):
				self._widget.insert('end', msg.get('notice/presence')
					+ ': ', tag)
			self._widget.mark_set(signer_mark_begin, 'end - 1 chars')
			self._widget.mark_gravity(signer_mark_begin, 'left')
			self._widget.insert('end',
				msg.get_signer('*verifying*'), tag)
			if msg.get('message/sender', None):
				self._widget.insert('end', " (%s)" %
					msg.get('message/sender'), tag)
			self._widget.insert('end', '\n', tag)
		# Format regular puff
		else:
			self._widget.insert('end', '\n', ('hrspacing', tag))
			self._widget.insert('end', '\n', ('hr', tag))
			self._widget.insert('end', '\n', ('hrspacing', tag))

			tags = ('bold', tag,)
			signer = msg.get_signer(None)
			if signer is None or (signer != '*unverified*' and\
				signer != '*unsigned*'):
				# Display a thumbnail pic
				if Config.SHOW_THUMBS:
					thumb_mark = 'thumb-mark-%i' % hash(msg)
					self._widget.mark_set(thumb_mark, 'end -1 chars')
					self._widget.mark_gravity(thumb_mark, 'left')
					if signer:
						url = msg.get('message/image', None)
						if type(url) not in STRING_TYPES:
							url = None
						thumbcache.getthumb(signer, url,
							lambda img, h=hash(msg), s=self:
							s.display_thumb(img, h),
							fromcache=not new)

				# Colorize sender attribute
				if signer:
					for key in Config.COLORS.keys():
						if re.search(key, signer):
							tags = tags + ('%s-fg' %
								Config.COLORS[key],)
							break

			self._widget.insert('end',
				msg.get('message/sender', 'Unknown sender'), tags)
			self._widget.insert('end',
				' [%s]' % msg.get_cat(), tag)
			if msg.get('message/subject', None) is not None:
				self._widget.insert('end',
					' "%s"' % msg.get('message/subject'), tag)
			if msg.get('id/class', None) is not None:
				self._widget.insert('end',
					' (%s)' % msg.get('id/class'), tag)
			msgstart = self._widget.index('end')
			self._widget.insert('end', '\n', tag)
			# Check for WQP in bodytext
			bodylines = string.split(bodytext, '\n')
			for i in range(len(bodylines)):
				line = bodylines[i]
				# Add the newline back to every line but the last
				if i != len(bodylines) - 1:
					line = line + '\n'
				if line[:2] == '> ':
					self._widget.insert('end', line,
						(tag, 'quotation'))
				else:
					self._widget.insert('end', line, tag)
			self._widget.insert('end', '-- ', (tag, 'right'))
			self._widget.mark_set(signer_mark_begin, 'end -1 chars')
			self._widget.mark_gravity(signer_mark_begin, 'left')
			self._widget.insert('end', msg.get_signer('*verifying*'), 
				(tag, 'right'))
			if msg.get_recipients():
				self._widget.insert('end', ' for %s' %
					string.join(msg.get_recipients(), ', '),
					(tag, 'right'))
			sentstr = ''
			if msg.get('id/time', None) is not None:
				sentstr = ' at %s\n' %\
					time.strftime("%m-%d %H:%M:%S",
					time.localtime(msg.get('id/time')))
			else:
				sentstr = '\n'
			self._widget.insert('end', sentstr, (tag, 'right'))

			# Check for URLs (in non-receipt puffs only)
			urlpat = re.compile(URLPAT)
			rowoffset = 0
			for line in string.split(bodytext, '\n'):
				coloffset = 0
				while 1:
					match = urlpat.search(line, coloffset)
					if match:
						(patstart, patend) = match.span()
						while line[patend-1] in ['.', ',']:
							patend = patend - 1
						coloffset = patend
						self._widget.tag_add('URL',
							'%s + %i lines + %i chars' %
							(msgstart, rowoffset, patstart),
							'%s + %i lines + %i chars' %
							(msgstart, rowoffset, patend))
					else:
						break
				rowoffset = rowoffset + 1


		# Set this to be the current puff on mouse click
		# This causes loads of problems with the other bindings
		self._widget.tag_bind(tag, '<Button-1>',
			lambda e, s=self, p=msg: s.set_curpuff_frompuff(p))
		self._widget.tag_bind(tag, '<Button-3>',
			lambda e, s=self, p=msg: s.set_curpuff_frompuff(p))

		# Delete old puffs from buffer
		if len(self._pufflist) > Config.SCROLLBACK:
			numdel = len(self._pufflist) - Config.SCROLLBACK
			to_be_deleted = self._pufflist[:numdel]
			# Compute start using the pufftag here, because the pufftags
			# will be deleted in the loop below
			first_remaining = self._pufflist[numdel]
			start = self._widget.tag_ranges(
				'pufftag-%i' % hash(first_remaining))[0]
			self._pufflist = self._pufflist[numdel:]
			for p in to_be_deleted:
				signer_mark_begin = 'signer-mark-%i-begin' % hash(p)
				thumb_mark = 'thumb-mark-%i' % hash(p)
				self._widget.mark_unset(signer_mark_begin)
				self._widget.mark_unset(thumb_mark)
				pufftag = 'pufftag-%i' % hash(p)
				self._widget.tag_delete(pufftag)
				# also delete URL tags
				self._widget.tag_remove('URL', 1.0, start)
			self._widget.delete(1.0, start)
			if self._curpuff and self._curpuff >= numdel:
				self._curpuff = self._curpuff - numdel

		# Only scroll at end if end is visible
		# Warning: do this after setting the current puff, above, since
		# set_curpuff will cause the text widget to scroll there
		if scroll_bottom:
			self._widget.see('end')
#			retval = self._widget.tag_nextrange(tag, 1.0)
#			if retval:
#				self._widget.see(retval[1])
#				self._widget.see(retval[0])

		self._widget.configure(text_state = 'disabled')
	
		if not new: return
		# Beep if it's encrypted
		if self._beep:
			self._master.bell()
		if self._command is not None:
			# TODO: set envvars like for gsubrc
			# TODO: figure out how to fork() on Win32
			if sys.platform != 'win32':
				args = string.split(self._command)
				args = [args[0]] + args
				pid = os.fork()
				if pid:
					# parent
					pass
				else:
					try:
						apply(os.execlp, tuple(args))
					except:
						sys.exit(0)
		if self._lift:
			self._master.switch_screen(Config.SUBLIST.index(self))
		if self._deiconify:
			self._master.master.deiconify()
		# Flash interacts weirdly with lifting the screen
		if self._highlight and not self._master.is_curscreen(self):
			index = Config.SUBLIST.index(self)
			self._master._buttonlist[index]['background'] = 'maroon'

	def display_thumb(self, thumb, hash):
		if thumb is not None:
			if not hasattr(self._widget, 'image_create'):
				self._master.gale_error('Unable to display thumbnail; '+
					'old version of Python/Tkinter?')
				return
			mark = 'thumb-mark-%i' % hash
			tag = 'pufftag-%i' % hash
			curstate = self._widget.component('text').configure('state')[-1]
			self._widget.component('text').configure(state='normal')
			try:
				index = self._widget.index(mark)
				self._widget.insert(index, ' ', tag)
				self._widget.image_create(index, image=thumb,
					align='bottom')
			except TclError, e:
				self._master.gale_error('Error displaying thumbnail: %s'
					% `e`)
			self._widget.component('text').configure(state=curstate)

			# Try to make the end of this puff visible if it's not
			# This may have weird effects
			ranges = self._widget.tag_ranges(tag)
			start = ranges[0]
			end = ranges[-1]
			if self._widget.dlineinfo(start):
				self._widget.see(end)
		# clear out unused thumbs in cache
		imagelist = []
		for c in Config.SUBLIST:
			image_str = c.widget().image_names()
			l = string.split(image_str)
			l = map(lambda x: string.split(x, '#')[0], l)
			imagelist = imagelist + l
		thumbcache.GC(imagelist)

	def puff_verified(self, puff):
		signer = puff.get_signer(None)
		if signer is None:
			return
		signer_mark_begin = 'signer-mark-%i-begin' % hash(puff)
		self._widget.configure(text_state = 'normal')

		# Display a thumbnail pic
		thumb_mark = 'thumb-mark-%i' % hash(puff)
		if Config.SHOW_THUMBS and thumb_mark in self._widget.mark_names()\
			and signer != '*unverified*' and signer != '*unsigned*':
			# Eeek, I don't like that *unverified* test
			thumbcache.getthumb(signer, None, lambda img, s=self,
				h=hash(puff): s.display_thumb(img, h))
		# Redraw signer
		try:
			index = self._widget.index(signer_mark_begin)
			self._widget.delete(signer_mark_begin, 
				'%s + 2 chars wordend wordend' % index)
			self._widget.insert(signer_mark_begin, signer)
		except TclError:
			pass
		self._widget.configure(text_state = 'disabled')

	def cancel_highlight(self):
		index = Config.SUBLIST.index(self)
		self._master._buttonlist[index]['background'] = self._origbg
	
	def clear_puffs(self):
		self._pufflist = []
		self._widget.configure(text_state = 'normal')
		self._widget.delete(1.0, 'end')
		marks = self._widget.mark_names()
		for name in marks:
			self._widget.mark_unset(name)
		pufftags = self._widget.tag_names()
		prefixlen = len('pufftag')
		for name in pufftags:
			if name[:prefixlen] == 'pufftag':
				self._widget.tag_delete(name)
		self._widget.configure(text_state = 'disabled')
		self._curpuff = None

