#!/uns/bin/python

import os, sys, regex, string, re, time
import pygale, gale_env, userinfo

#------------------------------------------------------------
# Global configuration
#------------------------------------------------------------
# Colorize puffs from different people
if sys.platform == 'win32':
	USE_COLOR = 0
else:
	USE_COLOR = 1
PEOPLE = {
	'egnor@' :		('blue', 'white'),
	'tlau@' :		('yellow', 'black'),
	'corin@' :		('green', 'black')
}

def makebold(text):
	if sys.platform == 'win32':
		return text
	else:
		return "[1m" + text + "[0m"

# Colors we know about
g_COLORS = {	'black' : 0,
			'red' : 1,
			'green' : 2,
			'yellow' : 3,
			'blue' : 4,
			'magenta' : 5,
			'cyan' : 6,
			'white' : 7
		}
def fgcode(color):
	return '3%i' % g_COLORS[color]
def bgcode(color):
	return '4%i' % g_COLORS[color]
def color(text, fg, bg, bold=1):
	if USE_COLOR:
		if bold:
			boldstr = '1;'
		else:
			boldstr = ''
		start = '[%s%s;%sm' % (boldstr, fgcode(fg), bgcode(bg))
		end = '[0m'
		return start + text + end
	else:
		return makebold(text)

def format(text):
	lines = string.split(text, '\n')
	out = []
	while lines:
		l = lines[0]
		lines = lines[1:]
		if len(l) > 80:
			wrappoint = string.rfind(l, ' ', 0, 80)
			if wrappoint == -1:
				# No space, don't wrap
				out.append(l)
			else:
				out.append(l[:wrappoint])
				lines.insert(0, l[wrappoint+1:])
		else:
			out.append(l)
	return string.join(out, '\n')

def show(puff):
	#--------------------------------------------------
	# Catch all the message headers in the envvars
	#--------------------------------------------------
	# Sender: real name of person sending puff
	sender = puff.get('message/sender', 'Unknown sender')
	# Category, eg "pub/comp/linux"
	category = puff.get_cat('Unknown category')
	# Sign: Gale ID of sender, eg "tlau@ofb.net"
	sign = puff.get_signer('Unknown sig')
	if sign is None:
		sign = 'Unknown sig'
	# Date message was sent, eg "1998-08-24 15:18:47"
	date = puff.get('id/time', None)
	if date is None:
		date = time.strftime('%m-%d %H:%M:%S', time.localtime(time.time()))
	else:
		date = time.strftime('%m-%d %H:%M:%S', time.localtime(date))
	# Recipient, plain text, eg "Tessa Lau"
	recipient = puff.get('message/recipient', '*everyone*')
	# Status of recipient for a receipt, eg "in/present"
	status = puff.get('notice/presence', 'Unknown status')
	# Client used to send puff
	idclass = puff.get('id/class', 'Unknown client')

	# Figure out who I am
	my_id = pygale.expand_id(gale_env.get('GALE_USER'))

	# Get the text of the message
	text = puff.get('message/body')
	if text is not None:
		text = format(text)

	# Filter out pings
	cats = string.split(category, ':')
	if '/ping' in cats:
		return

	# Receipts
	if regex.search('/receipt', category) != -1:
		s = '* %s received: %s <%s> (%s) via %s' %\
			(date, status, sign, sender, idclass)
		print s
		return
	elif regex.search('/notice/\w+/log', category) != -1:
		s = '* %s: %s <%s> (%s) via %s' %\
			(date, status, sign, sender, idclass)
		print s
		return

	if recipient == my_id:
		sys.stdout.write('\007')

	# Format message
	print '\r' + ('-' * 79)

	# Color-code the header by sender
	bold_sender = makebold(sender)
	for key in PEOPLE.keys():
		if sign is not None and re.match(key, sign) is not None:
			bold_sender = apply(color, (sender,) + PEOPLE[key])

	print "[%s] via %s\n%s" % (makebold(category), idclass, bold_sender)
	if text is not None:
		print text,
	bolded_sig = "-- <%s> for %s at %s --" %\
		(makebold(sign), makebold(recipient), date)
	normal_sig = "-- <%s> for %s at %s --" %\
		(sign, recipient, date)
	bolded_sig = ' ' * (79-len(normal_sig)) + bolded_sig
	print bolded_sig

if __name__ == '__main__':
	main()
