#!/usr/bin/env python

from Tkinter import *
import tkMessageBox
try:
	import tkFont
	HAVE_FONTS = 1
except ImportError:
	HAVE_FONTS = 0

import tkColorChooser
from types import *
import re, string, time, cPickle, os, sys, signal, getopt
import traceback

import Pmw
import buttonbar	# Peter's buttonbar
import search, thumbcache		# dialogs
import config, connection	# Connections to the gale server
from HistoryCombo import *
from pygale import *

# Default fonts
if sys.platform == 'win32':
	BOLD_FONT = 'Helvetica 10 bold'
	HR_FONT = 'SmallFont 2'
	HR_SPACING_FONT = 'SmallFont 2'
	HR_BACKGROUND = ''
	HR_RELIEF = 'raised'
	TEXT_FONT = 'Helvetica 10'
	QUOTE_FONT = 'Helvetica 10 italic'
else:
	BOLD_FONT = 'Times -14 bold'
	TEXT_FONT = 'Courier -12'
	QUOTE_FONT = '-*-courier-medium-o-*-*-12-*-*-*-*-*-*-*'
	HR_FONT = 'fixed -3'
	HR_SPACING_FONT = 'fixed -4'
	HR_BACKGROUND = ''
	HR_RELIEF = 'sunken'

# Categories to strip from a reply
CATSEPCHAR = '[\./]'
STRIPCATLIST = ['/q', '/url', '/fwd', '/bot']
STRIPCATS = string.join(map(lambda x, c=CATSEPCHAR: '%s$|%s%s' %
	(x, x, c), STRIPCATLIST), '|')

# Number of lines to save in log
LOG_MESG_BUFFER = 100

# Default presence strings
DEFAULT_PRESENCES = ['in/perhaps', 'out/somewhere/else']

# Whether or not to map the debug menu
DEBUG = 0

# Return an id/instance string
def getinstance():
	import socket
	dom = pygale.gale_env.get('GALE_DOMAIN', 'unknown domain')
	host = socket.gethostname()
	user = pygale.gale_env.get('GALE_USER', 'unknown user')
	if sys.platform == 'win32':
		display = 'Windows'
	else:
		display = pygale.gale_env.get('DISPLAY', 'unknown display')
	pid = str(os.getpid())
	return '%(dom)s/%(host)s/%(user)s/%(display)s/%(pid)s' % locals()

# Actually send this puff
# Returns 0 on success, something else otherwise
class PuffSender:
	def __init__(self, galewin, pubcat, privcat, subj, pufftext,
		callback = None):
		self.galewin = galewin
		self.puff = pygale.Puff()
		self.cat = []
		self.privcat = privcat
		self.pubcat = pubcat
		self.subj = subj
		self.pufftext = pufftext
		if pubcat:
			self.cat.append(pubcat)
		self.bad_ids = []
		self.recp_names = []
		self.recp_ids = []
		self.callback = callback

	def collect_ids(self, req_id, id_info):
		id, fullname = id_info
		if id is None or fullname is None:
			self.bad_ids.append(req_id)
		else:
			self.cat.append(pygale.id_category(id, 'user', ''))
			self.recp_names.append(fullname)
			self.recp_ids.append(id)
		if len(self.recp_ids) + len(self.bad_ids) == self.num_priv_recips:
			# All ids collected
			if not self.pubcat:
				self.puff.set_text('message/recipient', 
								   string.join(self.recp_names, ','))
			self.send_puff()

	def send(self):
		if self.privcat:
			# Filter out null ids
			privcats = map(pygale.expand_id, filter(None,
				re.split('[: ,]+', self.privcat)))
			if not self.pubcat and Config.CC_SELF and \
				Config.SIGNER not in privcats:
				# if it's a private puff and ccself is set
				privcats.append(Config.SIGNER)
			# Remove duplicates in list of privcats
			d = {}
			for x in privcats:
				d[x] = None
			privcats = d.keys()

			# Collects IDs
			self.num_priv_recips = len(privcats)
			for p in privcats:
				pygale.lookup_id(p, lambda i, s=self, p=p:
					s.collect_ids(p, i))
		else:
			return self.send_puff()
	
	def send_puff(self):
		if self.bad_ids:
			if len(self.bad_ids) == 1:
				msg = 'Cannot find user ' + self.bad_ids[0]
			else:
				msg='Cannot find users:\n' + string.join(self.bad_ids, '\n')
			m = tkMessageBox.showerror(
				title='Error: recipient list',
				parent=self.galewin,
				message=msg)
			if self.callback:
				self.callback(1)
			return 1

		self.puff.set_time('id/time', int(time.time()))
		self.puff.set_text('id/instance', getinstance())
		if Config.SENDER:
			self.puff.set_text('message/sender', Config.SENDER)
		if Config.THUMBNAIL_URL:
			self.puff.set_text('message/image', Config.THUMBNAIL_URL)
		self.puff.set_text('id/class', FUGU_VERSION_LONG)
		self.pufftext = re.sub('\n', '\r\n', self.pufftext)
		self.puff.set_text('message/body', self.pufftext)
		self.puff.set_cat(string.join(self.cat, ':'))
		if self.subj:
			self.puff.set_text('message/subject', self.subj)

		# Ask for a return receipt
		# TODO: there should be a way to turn this off
		if not self.pubcat:
			self.puff.set_text('question/receipt',
				pygale.id_category(Config.SIGNER, 'user', 'receipt'))

		# Sign
		if Config.SIGNER:
			out = self.puff.sign_message(Config.SIGNER)
			if out is None:
				# Uh oh, can't sign message
				if self.callback:
					self.callback(1)
				return 1
		else:
			out = self.puff

		# Encrypt
		if not self.pubcat:
			# Encrypt if being sent to individuals
			out.encrypt_message(self.recp_ids, lambda puff, s=self:
				s.finish_send(puff))
			self.puff.set_recipients(self.recp_ids)
		else:
			self.finish_send(out)
	
	def finish_send(self, out):
		try:
			self.galewin._galeconn.transmit_puff(out)
		except pygale.PyGaleErr, e:
			# server connection died
			self.galewin.gale_error(e)
			# Assume retry is already in progress
			return
		# Log copy of puffs as requested
		for c in Config.SUBLIST:
			log = c['logsent']
			if (log >= 1 and not self.pubcat) or log == 2:
				c.addpuff(self.puff, new=0)
		if self.callback:
			self.callback(0)
		return 0

def waitforzombie(*args):
	try:
		os.wait()
	except:
		pass

# Transition from old-style to new-style config class
class ConfigClass: pass
class Connection:
	def __getitem__(self, key):
		uname = '_' + key
		return self.__dict__[uname]

# ----------------------------------------------------------------------
# Global configuration object
# (this has to come after the declaration of Connection, above)
#class ConfigClass:
#	def __init__(self):
#		# Configuration
#		self.SCROLLBACK = 100		# Number of puffs to keep in window
#
#
#		# Presence strings
#		self.PRESENCE_STRINGS = DEFAULT_PRESENCES
#		# Default presence state
#		self.PRESENCE = self.PRESENCE_STRINGS[0]
#		self.STEALTH_MODE = 0
#		# Custom sender coloring
#		self.COLORS = {
#		}
#		# Puff sender
#		self.SENDER = get_my_fullname()
#		# Command to load a URL
#		self.URLLOAD = 'netscape -remote openURL(%s)'
#		# Puff highlight color
#		self.HIGHLIGHT = '#00ffff'
#		# Gale id to sign puffs with
#		self.SIGNER = pygale.gale_user()
#		# CC self on private puffs
#		self.CC_SELF = 0
#		# List of subscriptions
#		self.SUBLIST = [
#			Connection('Public', gale_env.get('GALE_GSUB', 'pub')),
#			Connection('Private', gale_env.get('GALE_SUBS',
#				pygale.id_category(pygale.gale_user(), 'user', ''))),
#			Connection('Watch', pygale.dom_category(pygale.gale_domain(),
#				'notice'))
#		]
#		# Whether to convert delimiters (JTR)
#		self.JIHAD_FLAG = 0
#		self.JIHAD_DELIMITER = '.'
#		# Whether or not to display puffbar buttons
#		self.MAP_PUFFBUTTONS = 1
#		# Whether or not to draw thumbnails (on by default only if your
#		# Tkinter supports it)
#		if hasattr(Text, 'image_create'):
#			self.SHOW_THUMBS = 1
#		else:
#			self.SHOW_THUMBS = 0
#		self.BIG_THUMBS = 0
#
#
#		# Window sizes
#		self.PANE_HEIGHT = 700
#		self.PANE_TOP = 500
#		self.PANE_BOTTOM = 200
#		self.PANE_WIDTH = -1
#		self.SAVE_PANE_WIDTH = 1

# ----------------------------------------------------------------------
# Handler for SIGUSR1
# Dump the contents of our screens into ~/.gale/fugu-<screen#>.dump
# For rescue/debug purposes only
def handle_usr1(*args):
	print 'Dumping contents of screens to', FUGUDIR
	for i in range(len(Config.SUBLIST)):
		f = open('%s/fugu-%i.dump' % (FUGUDIR, i), 'w')
		text = string.replace(
			Config.SUBLIST[i].widget().get(1.0, 'end'),
			'\n\n\n', '\n' + ('-' * 79))
		f.write(text)
		f.close()

class GaleWin(Frame):
	def __init__(self, master=None):
		Frame.__init__(self, master)
		self.master.title('Fugu')
		self.master.withdraw()
		self.master.protocol('WM_DELETE_WINDOW', self.quit)
		self.pack(expand=1, fill=BOTH)

		self._tagid = 0			# sequence number for tags
		self._curscreen = None		# currently selected screen
		self._buttonlist = []		# List of screen buttons
		self._clipboard = None			# currently selected url
		self._thumbs = {}			# thumbnail pics
		self._thumb_cbs = {}		# callbacks on thumbnails

		# Load global configuration object
		# Be sure to do this before creating widgets, below
		#self.load_config()

		self._create_widgets()
		self._bind_hotkeys()

		# Set up an error handler
		pygale.set_error_handler(self.gale_error)
		pygale.set_update_handler(self.gale_error)
		# Set up SIGUSR1 handler
		if sys.platform != 'win32':
			signal.signal(signal.SIGUSR1, handle_usr1)
			signal.signal(signal.SIGCHLD, waitforzombie)
			signal.signal(signal.SIGPIPE, self.quit)
		signal.signal(signal.SIGTERM, self.quit)
		signal.signal(signal.SIGINT, self.quit)

		# Set up writeable gale connection and presence state
		self.make_send_conn(reconnect=0)
		
		# Create screen popup
		self.screenpop = ScreenPop(self)
		self.screenpop.withdraw()

		# Create log window
		self.logwin = LogWin(self)
		self.logwin.withdraw()

		# Set us up with the public category in focus
		self._puffwidget.set_pub_focus()
		self.master.deiconify()

		# Set up icon
		for dir in sys.path:
			if dir == "": dir = '.'
			try:
				self.master.iconbitmap("@%s/fugu-icon.xbm" % dir)
				break
			except TclError:
				continue

		# Persist pufflog every 5 minutes
		if Config.PERSIST_INTERVAL:
			self.after(Config.PERSIST_INTERVAL*1000, self.persist_pufflog)
	
	# Persist the pufflog
	def persist_pufflog(self, *args):
		try:
			Config.persist_pufflog(FUGUDIR)
		except:
			pass
		if Config.PERSIST_INTERVAL:
			self.after(Config.PERSIST_INTERVAL*1000, self.persist_pufflog)
	
	# Clear out the presence box
	def clear_presences(self):
		Config.PRESENCE_STRINGS = ['in/perhaps']
		Config.DEFAULT_PRESENCE = Config.PRESENCE_STRINGS[0]
		self.presence_box.component('scrolledlist').setlist(
			Config.PRESENCE_STRINGS)
		self.presence_box.selectitem(0)
		self.change_presence()

	# Invoke and focus the presence box
	def drop_presence(self, *args):
		self.presence_box.invoke()
		self.presence_box.component('listbox').focus_set()

	def change_presence(self, in_presence = None):
		if in_presence is None:
			in_presence = self.presence_box.get(first=None)
		if Config.BROADCAST_PRESENCE:
			while 1:
				try:
					pygale.notify_in(self._galeconn,
						Config.SIGNER, in_presence,
						version=FUGU_VERSION_LONG,
						instance=getinstance())
					pygale.notify_out(self._galeconn, Config.SIGNER,
						'out/quit', version=FUGU_VERSION_LONG,
						instance=getinstance())
				except pygale.PyGaleErr:
					self.make_send_conn()
				else:
					break
	
	def make_send_conn(self, reconnect=1):
		self._galeconn = pygale.GaleClient()
		self._galeconn.set_onconnect(lambda h, r=reconnect, s=self:
			s.on_sendconn(h, r))
		self._galeconn.set_ondisconnect(self.on_disconnect)
		self._galeconn.connect(self.make_send_conn2)
	
	def make_send_conn2(self, host):
		if not host:
			self.gale_error('Unable to connect to a gale server')
			return
		self.gale_error('Connected to server at %s' % host)

	def on_sendconn(self, host, reconnect):
		# Set up AKD server
		akd_cat = pygale.id_category(Config.SIGNER, 'auth/query')
		self._galeconn.sub_to(akd_cat)
		self._galeconn.set_puff_callback(lambda p, s=self:
			s.akd_request(p))

		# Transmit login notice
		if reconnect:
			self.change_presence('in/reconnected')
		else:
			self.change_presence()

	def on_disconnect(self):
		self.gale_error('Disconnected from server')
	
	def gale_error(self, msg):
		date = time.strftime('%m-%d %H:%M:%S ',
			time.localtime(time.time()))
		msg = date + string.strip(msg)
		self.set_status(msg)
		# TODO: 1) fix hasattr ickiness
		if hasattr(self, 'logwin'):
			self.logwin.add_message(msg)

	def _create_widgets(self):
		# Menus
		menubar = Frame(self, relief='flat', borderwidth=0)
		menubar.pack(side=TOP, fill=X, expand=0)
		self.menubar = menubar

		filemenu = Menubutton(menubar, text="File", underline=0)
		filemenu.pack(side=LEFT)
		filemenu.menu = Menu(filemenu, tearoff=0)

		filemenu.menu.add_command(label='Reconnect to server',
			underline=0, command=self.reconnect_all)
		filemenu.menu.add_command(label='Reset presence string menu',
			underline=0, command=self.clear_presences)
			
		filemenu.menu.add_separator()
		filemenu.menu.add_command(label='Preferences',
			underline=0, command=self.config_pop)
		filemenu.menu.add_command(label='Exit',
			underline=1, command=self.quit)
		filemenu['menu'] = filemenu.menu

		puffmenu = Menubutton(menubar, text="Puff", underline=0)
		puffmenu.pack(side=LEFT)
		puffmenu.menu = Menu(puffmenu, tearoff=0)
		puffmenu.menu.add_command(label='Clear puff (A-c)',
			underline=0, command=self.clearpuff)
		puffmenu.menu.add_command(label='Send puff (C-Enter)',
			underline=0, command=self.sendbutton)
		puffmenu.menu.add_command(label='Detach puff (A-d)',
			underline=0, command=self.detach)
		puffmenu.menu.add_separator()
		puffmenu.menu.add_command(label='Reply to puff (C-r)',
			underline=0, command=self.reply)
		puffmenu.menu.add_command(label='Reply to sender (C-R)',
			command=self.reply_to_sender)
		puffmenu.menu.add_command(label='Options (C-o)',
			underline=0, command=self.optbutton, state='disabled')
		puffmenu.menu.add_separator()
		puffmenu.menu.add_command(label='Previous in thread (C-p)',
			underline=0, command=self.prev_thread)
		puffmenu.menu.add_command(label='Next in thread (C-n)',
			underline=0, command=self.next_thread)
		puffmenu.menu.add_command(label='Author search (A-a)',
			underline=0, command=self.author_search)
		# For debugging
		if DEBUG:
			puffmenu.menu.add_separator()
			puffmenu.menu.add_command(label='Show marks',
				command=self.show_marks)
			puffmenu.menu.add_command(label='Show tags',
				command=self.show_tags)

		puffmenu['menu'] = puffmenu.menu

		self.screenmenu = Menubutton(menubar, text="Screens", underline=0)
		self.screenmenu.pack(side=LEFT)
		self.screenmenu.menu = Menu(self.screenmenu, tearoff=0)
		self.screenmenu.menu.add_command(label='Search (C-s)', underline=0,
			command=self.search_pop)
		self.screenmenu.menu.add_command(label='Configure',
			underline=0, command=self.screen_pop)
		self.screenmenu.menu.add_command(label='Clear current',
			underline=1, command=self.clear_curscreen)
		self.screenmenu.menu.add_separator()
		# Populate the screen menu
		self.regen_screenmenu()

		self.screenmenu['menu'] = self.screenmenu.menu

		if DEBUG:
			debugmenu = Menubutton(menubar, text='Debug')
			debugmenu.pack(side=LEFT)
			debugmenu.menu = Menu(debugmenu, tearoff=0)
			debugmenu.menu.add_command(label='GC', underline=0,
				command=self.garbage_collect)
			self.var_d_asyncurl = BooleanVar()
			self.var_d_asyncurl.set(asyncurl.DEBUG)
			self.var_d_authcache = BooleanVar()
			self.var_d_authcache.set(authcache.DEBUG)
			self.var_d_engine = BooleanVar()
			self.var_d_engine.set(engine.DEBUG)
			self.var_d_pygale = BooleanVar()
			self.var_d_pygale.set(pygale.DEBUG)
			self.var_d_sign = BooleanVar()
			self.var_d_sign.set(sign.DEBUG)
			self.var_d_thumbcache = BooleanVar()
			self.var_d_thumbcache.set(thumbcache.DEBUG)
			debugmenu.menu.add_checkbutton(label='Asyncurl', underline=0,
				variable=self.var_d_asyncurl,
				command=lambda s=self, v=self.var_d_asyncurl,
				n='asyncurl.DEBUG':
				s.toggle_debug_var(v, n))
			debugmenu.menu.add_checkbutton(label='Authcache', underline=0,
				variable=self.var_d_authcache,
				command=lambda s=self, v=self.var_d_authcache,
				n='authcache.DEBUG':
				s.toggle_debug_var(v, n))
			debugmenu.menu.add_checkbutton(label='Engine', underline=0,
				variable=self.var_d_engine,
				command=lambda s=self, v=self.var_d_engine,
				n='engine.DEBUG':
				s.toggle_debug_var(v, n))
			debugmenu.menu.add_checkbutton(label='Pygale', underline=0,
				variable=self.var_d_pygale,
				command=lambda s=self, v=self.var_d_pygale,
				n='pygale.DEBUG':
				s.toggle_debug_var(v, n))
			debugmenu.menu.add_checkbutton(label='Sign', underline=0,
				variable=self.var_d_sign,
				command=lambda s=self, v=self.var_d_sign,
				n='sign.DEBUG':
				s.toggle_debug_var(v, n))
			debugmenu.menu.add_checkbutton(label='Thumbcache', underline=0,
				variable=self.var_d_thumbcache,
				command=lambda s=self, v=self.var_d_thumbcache,
				n='thumbcache.DEBUG':
				s.toggle_debug_var(v, n))
			debugmenu['menu'] = debugmenu.menu

		helpmenu = Menubutton(menubar, text='Help', underline=0)
		helpmenu.pack(side=RIGHT)
		helpmenu.menu = Menu(helpmenu, tearoff=0)
		helpmenu.menu.add_command(label='Error messages',
			underline=0, command=self.popup_logwin)
		helpmenu.menu.add_command(label='About',
			underline=0, command=self.about)
		helpmenu['menu'] = helpmenu.menu

		# Presence status dropdown
		self.presence_box = Pmw.ComboBox(menubar, scrolledlist_items=
			Config.PRESENCE_STRINGS,
			selectioncommand=self.change_presence,
			entry_takefocus=0, arrowbutton_takefocus=0)
		self.presence_box.selectitem(Config.DEFAULT_PRESENCE)
		self.presence_box.pack(side=RIGHT, expand=0, fill=BOTH)
		self.presence_box.component('entry').bind('<Key-Up>',
			self.drop_presence)
#		self.presence_box.component('entry').bind('<Key-Up>',
#			self.ret_break, add=1)
		self.presence_box.component('entry').bind('<Key-Down>',
			self.drop_presence)
#		self.presence_box.component('entry').bind('<Key-Down>',
#			self.ret_break, add=1)
		# These two don't work yet
		self.presence_box.component('entry').bind('<Return>',
			self.tabForward, add=1)
		#self.presence_box.component('listbox').bind('<Return>',
		#	self.tabForward, add=1)

		# Button bar
		self.buttonbar = buttonbar.ButtonBar(self)
		self.buttonbar['cursor'] = 'top_left_arrow'
		self.buttonbar.pack(side=TOP, expand=0, fill=BOTH)

		for conf in Config.SUBLIST:
			button = self.add_button(conf)

		# Main text widget
		pane = Pmw.PanedWidget(self, hull_height=Config.PANE_HEIGHT)
		pane.add('bottom', min=0, size=Config.PANE_BOTTOM)
		pane.insert('top', before='bottom', min=0, size=Config.PANE_TOP)
		self.pane = pane
		# Pack the pane after creating its child text widgets
	
		# Status bar
		self._status = Label(pane.pane('bottom'), anchor='w')
		self._status.pack(side=BOTTOM, fill=X, expand=0, padx=2)

		# Button frame
		self.puffbuts = Frame(pane.pane('bottom'))
		self._sendbutton = Button(self.puffbuts, text='Send',
			command=self.sendbutton, takefocus=0, padx=2, pady=1)
		self._sendbutton.pack(side=LEFT)
		self._optbutton = Button(self.puffbuts, text='Options',
			command=self.optbutton, takefocus=0, state='disabled',
			pady=1, padx=2)
		self._optbutton.pack(side=LEFT)
		self._detachbutton = Button(self.puffbuts, text='Detach',
			command=self.detach, takefocus=0, pady=1, padx=2)
		self._detachbutton.pack(side=LEFT)
		if Config.MAP_PUFFBUTTONS:
			self.puffbuts.pack(side=TOP, fill=X, expand=0, padx=2)

		# --------------------------------------------------
		# Puff entry widgets

		self._puffwidget = PuffWidget(pane.pane('bottom'))
		self._puffwidget.pack(side=TOP, fill=BOTH, expand=1, padx=2,
			pady=2)

		# Right button menu
		# NOTE: when anything changes in this menu, also modify the
		# popup_context function
		self.context_menu = Menu(self, tearoff=0)
		self.context_menu.add_command(label='Reply (C-r)', command=self.reply,
			underline=0)
		self.context_menu.add_command(label='Reply to sender (C-R)',
			command=self.reply_to_sender, underline=9)
		self.context_menu.add_command(label='Next in thread (C-n)',
			command=self.next_thread, underline=0)
		self.context_menu.add_command(label='Previous in thread (C-p)',
			command=self.prev_thread, underline=0)
		self.context_menu.add_command(label='Author search (A-a)',
			command=self.author_search, underline=0)
		self.context_menu.add_command(label='Cite puff (A-t)',
			command=self.cite_puff, underline=2)
		self.context_menu.add_separator()
		self.context_menu.add_command(label='Copy URL to clipboard',
			command=self.own_selection, underline=1, state='disabled')
		self.context_menu.add_separator()
		self.context_menu.add_command(label='Unsub to this category',
			command=self.kill_cat, underline=0)
		self.context_menu.add_command(label='Detach to new window',
			underline=0, command=self.puffdisp)

		# The ordering here is tricky.  The addpuff() calls must happen
		# after the status bar has been created (above).	The
		# pane.configure must happen after the text widgets have been
		# created.
		for conf in Config.SUBLIST:
			text = self.make_text_widget()
			# Create a Gale connection
			# This should be done later, when the send connection is
			# created
			conf.setwidget(self, text)
			text.after_idle(lambda pl=conf.pufflist(), s=self, c=conf:
				s.idle_pufflist(c, pl))
			conf._pufflist = []

		self.switch_screen(0)

		# Configure the pane width after the text widget's created and
		# packed
		if not Config.SAVE_PANE_WIDTH or Config.PANE_WIDTH == -1:
			pane.update_idletasks()
			width = 0
			for name in pane.panes():
				w = pane.pane(name).winfo_reqwidth()
				if width < w:
					width = w
			pane.configure(hull_width = width)
		else:
			pane.configure(hull_width = Config.PANE_WIDTH)
		pane.pack(side=TOP, expand=1, fill=BOTH)

	# ----------------------------------------------------------------------
	# Debug menu
	def toggle_debug_var(self, boolvar, debugvar):
		val = boolvar.get()
		exec '%s = %i' % (debugvar, val)

	def garbage_collect(self, *args):
		if sys.version >= '2.0':
			import gc
			retval = gc.collect()
			print 'Collector: %i unreachables' % retval
		else:
			print 'Explicit GC not supported in Python < 2.0'
	# ----------------------------------------------------------------------
	
	def show_subject(self):
		self._puffwidget.show_subject()
	
	def hide_subject(self):
		self._puffwidget.hide_subject()

	def idle_pufflist(self, conf, pufflist):
		for p in pufflist:
			conf.addpuff(p, 0)
		conf.widget().see('end')
		conf.connect()

	# Make a text widget to contain puffs
	def make_text_widget(self):
		text = Pmw.ScrolledText(self.pane.pane('top'), text_height=50,
			text_width=81, text_selectforeground='white',
			text_selectbackground='black', text_selectborderwidth=0,
			text_wrap='word', text_font=TEXT_FONT,
			vertscrollbar_takefocus=0,
			vertscrollbar_width=10)
		self.set_text_cursor(text)

		text.component('text').bind('<Button-3>', self.url_store)
		text.component('text').bind('<Button-3>', self.popup_context, add=1)
		# Bind these here to override Tk's default binding for
		# Button4/Button5 in the text widget (in 8.3, but not in 8.0)
		text.component('text').bind('<Button-4>', self.page_up)
		text.component('text').bind('<Button-5>', self.page_down)

		# Tags for the text widget
		text.tag_config('bold', font=BOLD_FONT)

		# URLs
		text.tag_config('URL', foreground='blue',
			underline=1)
		text.tag_bind('URL', '<Enter>',
			lambda e, s=self, t=text: s.set_text_cursor(t, 'hand2'))
		text.tag_bind('URL', '<Leave>',
			lambda e, s=self, t=text: s.set_text_cursor(t))
		text.tag_bind('URL', '<Button-1>', self.url_store)
		text.tag_bind('URL', '<Button-1>', self.url_click, add=1)
		text.tag_config('URLclick', foreground='red',
			underline=1)
		# Lower it so its binding is executed first
		text.tag_lower('URL')

		# Tags for nifty horizontal rule
		text.tag_config('hr', font=HR_FONT, relief=HR_RELIEF,
			borderwidth=1, background=HR_BACKGROUND)
		text.tag_config('hrspacing', font=HR_SPACING_FONT)
		text.tag_config('right', justify='right')

		# Quotation
		if HAVE_FONTS:
			quotefont = tkFont.Font(self, QUOTE_FONT)
			quotewidth = quotefont.measure('> ') + 4
			text.tag_config('quotation', font=QUOTE_FONT,
				lmargin1=4, lmargin2=quotewidth, rmargin=4)
		else:
			# Just make a guess for how much to indent subsequent lines
			text.tag_config('quotation', font=QUOTE_FONT,
				lmargin1=4, lmargin2=10, rmargin=4)

		# Custom colors
		colors = Config.COLORS.values()
		for c in colors:
			text.tag_config('%s-fg' % c, foreground=c)

		text.configure(text_state = 'disabled')
		return text
	
	def add_button(self, conf):
		button = Button(self.buttonbar, text=conf['name'],
			highlightthickness=0, takefocus=0, pady=1, padx=2)
		button['command'] = lambda s=self, b=button:\
			s.switch_screen_unknown(b)
		self.buttonbar.add_button(button)
		self._buttonlist.append(button)
	
	def insert_button(self, button, index=None):
		if index is None:
			self._buttonlist.append(button)
		else:
			self._buttonlist.insert(index, button)
		self.buttonbar.add_button(button, index)
	
	def del_button(self, index):
		button = self._buttonlist[index]
		self.buttonbar.del_button(button)
		self._buttonlist[index:index+1] = []
		return button
	
	def change_button_name(self, index, name):
		self._buttonlist[index]['text'] = name

	# ------------------------------------------------------------
	# Reconnect screens
	def reconnect_all(self):
		for c in Config.SUBLIST:
			c.make_connection()
		self.make_send_conn()

	# ------------------------------------------------------------
	# Callbacks for ScreenPop
	def move_button_up(self, index):
		button = self.del_button(index)
		self.insert_button(button, index-1)

		# Old code
#		button = self._buttonlist[index]
#		button.forget()
#		print 'Forgetting button:', [button]
#		self._buttonlist[index:index+1] = []
#		print 'buttonlist after delete:', self._buttonlist
#		self._buttonlist.insert(index-1, button)
#		print 'buttonlist after insert:', self._buttonlist
#		if index == 1:
#			print 'Packing before:', [self._buttonlist[1]]
#			button.pack(before=self._buttonlist[1], side=LEFT,
#				expand=0, padx=4, pady=2)
#		else:
#			print 'Packing after:', [self._buttonlist[index-2]]
#			button.pack(after=self._buttonlist[index-2], side=LEFT,
#				expand=0, padx=4, pady=2)
		self.regen_screenmenu()

	def move_button_down(self, index):
		button = self.del_button(index)
		self.insert_button(button, index+1)

#		button = self._buttonlist[index]
#		button.forget()
#		self._buttonlist[index:index+1] = []
#		self._buttonlist.insert(index+1, button)
#		button.pack(after=self._buttonlist[index], side=LEFT,
#			expand=0, padx=4, pady=2)
		self.regen_screenmenu()

	# ------------------------------------------------------------

	# Switch to the requested screen
	# Index starts at 0
	def switch_screen(self, index):
		# Return immediately if this isn't a valid screen number
		if index >= len(Config.SUBLIST): return
		Config.SUBLIST[index].cancel_highlight()
		# Don't do anything else if we're already lifted
		if index == self._curscreen: return
		if self._curscreen == index: return
		# Raise old button
		if self._curscreen is not None:
			self._buttonlist[self._curscreen]['relief'] = 'raised'
			Config.SUBLIST[self._curscreen].widget().forget()
		Config.SUBLIST[index].widget().pack(side=TOP, expand=1,
			fill=BOTH, padx=2, pady=2)
		self._curscreen = index
		self._buttonlist[index]['relief'] = 'sunken'
	
	# Invoked on a button click
	def switch_screen_unknown(self, button):
		index = self._buttonlist.index(button)
		self.switch_screen(index)

	def is_curscreen(self, screen):
		return Config.SUBLIST[self._curscreen] == screen
	
	def _category_regexp(self, catstr):
		# Split catstring into list of cats, filter out condiments
		cats = filter(lambda x: x and (x[0] != '/'),
			string.split(catstr, ':'))
		# Split each cat on the intracategory delimiter
		cats = map(lambda x: re.split('[\./]', x), cats)
		# Join up each cat, escaping non-alphanumerics
		cats = map(lambda x: string.join(map(re.escape, x), '[\./]'), cats)
		return re.compile(string.join(cats, '|'))
	
	# Jump to the next/prev thread matching the current puff's category
	def next_thread(self, *args):
		client = Config.SUBLIST[self._curscreen]
		if client.curindex() is None: return
		orig_re = self._category_regexp(client.curpuff().get_cat())
		puffindex = client.curindex()
		for i in range(puffindex+1, len(client)):
			puffcats = string.split(client[i].get_cat(), ':')
			for d in puffcats:
				if orig_re.match(d):
					client.set_curpuff(i)
					return
		self.set_status('Last puff in this thread')

	def prev_thread(self, *args):
		client = Config.SUBLIST[self._curscreen]
		if client.curindex() is None: return
		orig_re = self._category_regexp(client.curpuff().get_cat())
		puffindex = client.curindex()
		for i in range(puffindex-1, -1, -1):
			puffcats = string.split(client[i].get_cat(), ':')
			for d in puffcats:
				if orig_re.match(d):
					client.set_curpuff(i)
					return
		self.set_status('First puff in this thread')

	# Find the next puff by this author
	def author_search(self, *args):
		client = Config.SUBLIST[self._curscreen]
		if client.curindex() is None: return
		puff = client.curpuff()
		sender = puff.get('message/sender', None)
		signer = puff.get_signer(None)
		puffindex = client.curindex()
		for i in range(puffindex+1, len(client)) +\
			range(0, puffindex):
			if i == 0 and puffindex != 0:
				self.set_status('Author search: wrapping to top')
			if sender and sender ==\
				client[i].get('message/sender', None):
				client.set_curpuff(i)
				return
			elif signer and signer ==\
				client[i].get_signer(None):
				client.set_curpuff(i)
				return
		self.set_status('No more puffs by this author')
	
	# Add a puff citation to the text widget
	def cite_puff(self, *args):
		puff = Config.SUBLIST[self._curscreen].curpuff()
		signer = puff.get_signer('anonymous')
		atpos = string.find(signer, '@')
		if atpos != -1:
			signer = signer[:atpos]
		pufftime = puff.get('id/time', None)
		if pufftime is None:
			timestr = ''
		else:
			timestr = time.strftime('%H:%M:%S',
				time.localtime(pufftime))
		self._clipboard = '%s[%s]' % (signer, timestr)
		self._puffwidget.text().insert('insert', self._clipboard, 'sel')
		self._puffwidget.text().tag_remove('sel', '1.0', 'end')
		self.own_selection()
	
	def puffdisp(self):
		import puffdisp
		puff = Config.SUBLIST[self._curscreen].curpuff()
		if puff is None: return
		p = puffdisp.PuffDisp(puff)
	
	# Pop up the right-button context menu
	# WARNING: the constant here has to be changed whenever you add
	# something to the context menu
	def popup_context(self, event):
		if self._clipboard is not None:
			self.context_menu.entryconfigure(7, state='normal')
		else:
			self.context_menu.entryconfigure(7, state='disabled')
		self.context_menu.tk_popup(event.x_root, event.y_root, 0)
	
	def own_selection(self):
		# Manipulate X selection
		self.selection_own()
		self.selection_handle(self.handle_selection)

	def handle_selection(self, start, end):
		if self._clipboard is not None:
			start = string.atoi(start)
			end = string.atoi(end)
			return self._clipboard[start:end]
		else:
			return ''

	# If a URL is under the cursor, store it in self._clipboard
	def url_store(self, event):
		client = Config.SUBLIST[self._curscreen]
		if 'URL' in client.widget().tag_names(
			"@%i,%i" % (event.x, event.y)):
			(start, end) = client.widget().tag_prevrange('URL',
				"@%i,%i" % (event.x, event.y))
			self._clipboard = client.widget().get(start, end)
		else:
			self._clipboard = None

	# Callback for clicking on a URL
	def url_click(self, event):
		if self._clipboard is None:
			return
		client = Config.SUBLIST[self._curscreen]
		(start, end) = client.widget().tag_prevrange('URL',
			"@%i,%i" % (event.x, event.y))
		client.widget().tag_add('URLclick', start, end)
		self.set_status('Loading url: %s' % self._clipboard)
		self.update_idletasks()
		if sys.platform == 'win32':
			try:
				import win32api
				url = re.sub(',', '%2C', self._clipboard)
				win32api.ShellExecute(0, None, url, None, '.', 0)
			except:
				self.gale_error('Unable to load URL (win32 ' +\
					'extensions not installed?)')
		else:
			try:
				urlcmd = Config.URLLOAD % re.sub(',', '%2C',
					self._clipboard)
			except TypeError, e:
				self.gale_error('Unable to construct URL command: ' +
					str(e))
				client.widget().tag_remove('URLclick', start, end)
				return
			if os.fork():
				# Parent
				client.widget().tag_remove('URLclick', start, end)
				return 'break'
			else:
				try:
					# Child
					args = string.split(urlcmd)
					args = [args[0]] + args
					apply(os.execlp, tuple(args))
				except:
					traceback.print_exc()
					sys.exit(0)

	# Kill these categories
	def kill_cat(self):
		client = Config.SUBLIST[self._curscreen]
		cat = client.curpuff().get_cat()
		negated_cats = map(lambda x: '-' + x, string.split(cat, ':'))
		negcat = string.join(negated_cats, ':')
		client['sub'] = client['sub'] + ':' + negcat
		client.reconnect()

	# What to do on a quit
	def quit(self, *args):
		Config.PANE_HEIGHT = self.pane.winfo_height()
		Config.PANE_TOP = self.pane.pane('top').winfo_height()
		Config.PANE_BOTTOM = self.pane.pane('bottom').winfo_height() + 2
		if Config.SAVE_PANE_WIDTH:
			Config.PANE_WIDTH = self.pane.pane('top').winfo_width()
		else:
			Config.PANE_WIDTH = -1

		# Save presence strings
		Config.PRESENCE_STRINGS = list(self.presence_box.get(0, 'end'))
		Config.DEFAULT_PRESENCE = self.presence_box.get()

		try:
			Config.save(FUGUDIR)
			Config.persist_pufflog(FUGUDIR)
		except:
			pass
		Frame.quit(self)
	
	def set_text_cursor(self, widget, cur='top_left_arrow'):
		widget.configure(text_cursor = cur)

	# Tab to next widget binding
	def tabForward(self, event):
		w = event.widget.tk_focusNext()
		w.focus_set()
		return 'break'
	
	# Paging up and down
	def page_up(self, event):
		conn = Config.SUBLIST[self._curscreen]
		conn.widget().yview('scroll', -1, 'pages')
		return 'break'

	def page_down(self, event):
		conn = Config.SUBLIST[self._curscreen]
		conn.widget().yview('scroll', 1, 'pages')
		return 'break'
	
	# This is more than just invoking the send button; it also has to
	# return 'break' so the return doesn't get processed and add a newline
	# to the message.
	def control_return(self, event):
		self.sendbutton()
		return 'break'

	# Binding script that returns break
	def ret_break(self, event):
		return 'break'

	# Define some nice keyboard shortcuts
	def _bind_hotkeys(self):
		# Hotkeys
		self._puffwidget.bind('<Control-Return>', self.control_return)

		self._puffwidget.bind('<Alt-c>', lambda e, s=self:
			s.clearpuff())
		self._puffwidget.bind('<Alt-d>', lambda e, s=self:
			s._detachbutton.invoke())
		self._puffwidget.bind('<Control-r>',
			lambda e, s=self: s.reply())
		self._puffwidget.bind('<Control-R>',
			lambda e, s=self: s.reply_to_sender())
		self._puffwidget.bind('<Key-Up>',
			lambda e, s=self: s.key_up(), text=0)
		self._puffwidget.bind('<Key-Down>',
			lambda e, s=self: s.key_down(), text=0)
		self._puffwidget.bind('<Alt-a>', self.author_search)
		self._puffwidget.bind('<Control-z>', lambda e, s=self:
			s.master.iconify())
		self._puffwidget.bind('<Control-s>', self.search_pop)
		self._puffwidget.bind('<Control-Prior>', self.home)
		self._puffwidget.bind('<Control-Next>', self.end)

		# Focus presence widget
		self._puffwidget.bind('<Control-Key-P>', lambda e, s=self:
			s.presence_box.component('entry').focus_set())

		# These bindings only work in the entry fields
		self._puffwidget.bind('<Control-n>', self.next_thread, text=0)
		self._puffwidget.bind('<Control-p>', self.prev_thread, text=0)
		self._puffwidget.bind('<Prior>', self.page_up, text=0)
		self._puffwidget.bind('<Next>', self.page_down, text=0)

		# Mouse button bindings have to be bound to the master frame
		self.master.bind('<Button-4>', self.page_up)
		self.master.bind('<Button-5>', self.page_down)

		# Alt-up/down always move the puff cursor
		self._puffwidget.bind('<Alt-Up>', self.key_up)
		self._puffwidget.bind('<Alt-Down>', self.key_down)
		self._puffwidget.bind('<Alt-Prior>', self.page_up)
		self._puffwidget.bind('<Alt-Next>', self.page_down)

		# Cut the reference to the cursor
		self._puffwidget.bind('<Alt-t>', self.cite_puff)

		# This next line causes problems; it binds to all entry widgets,
		# which it oughtn't do
		#self.master.bind_class('Entry', '<Return>', self.tabForward)

		# Break after certain commands in the text widget so it doesn't
		# change the puff selection
		# Perhaps this ought to bind only to the puff entry widget, and
		# not all text widgets
#		self.master.bind_class('Text', '<Up>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Down>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Shift-Up>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Shift-Down>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Control-a>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Control-n>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Control-p>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Prior>', self.ret_break, add=1)
#		self.master.bind_class('Text', '<Next>', self.ret_break, add=1)

		# Switching between screens with C-N where 1 <= N <= 9
		for i in range(1, 10):
			self._puffwidget.bind('<Control-Key-%i>' % i, lambda e, i=i,
				s=self: s.switch_screen(i-1))
	
	def home(self, *args):
		Config.SUBLIST[self._curscreen].widget().see(1.0)
	def end(self, *args):
		Config.SUBLIST[self._curscreen].widget().see('end')

	# Pop up a configuration dialog
	def config_pop(self):
		w = ConfigPop(self)
	
	# Pop up a search dialog
	def search_pop(self, *args):
		w = search.SearchDialog(self,
			Config.SUBLIST[self._curscreen].widget(),
			Config.SUBLIST[self._curscreen]['name'])

	def screen_pop(self):
		self.screenpop.deiconify()
		self.screenpop.set_screen(self._curscreen)
		self.screenpop.tkraise()

	def clear_curscreen(self):
		Config.SUBLIST[self._curscreen].clear_puffs()
	
	def show_marks(self):
		marks = Config.SUBLIST[self._curscreen].widget().mark_names()
		for name in marks:
			print name
		print 'Total:', len(marks)
	def show_tags(self):
		tags = Config.SUBLIST[self._curscreen].widget().tag_names()
		for name in tags:
			print name
		print 'Total:', len(tags)
	
	def regen_screenmenu(self):
		i = 0
		# WARNING: update this number when the contents of the screen menu
		# change!!!
		self.screenmenu.menu.delete(3, 'end')
		self.screenmenu.menu.add_separator()
		for conf in Config.SUBLIST:
			self.screenmenu.menu.add_command(label=conf['name'] +
				" (C-%i)" % (i+1),
				command=lambda s=self, i=i: s.switch_screen(i))
			i = i + 1
	
	# Actually send this puff
	# Returns 0 on success, something else otherwise
	def sendpuff(self, pubcat, privcat, subj, pufftext, callback = None):
		puff_sender = PuffSender(self, pubcat, privcat, subj, pufftext,
			callback)
		puff_sender.send()

	# What to do when we receive a puff on this connection
	def recvpuff(self, conn, puff):
		if puff.get('question/receipt', None) is not None:
			self.after_idle(lambda s=self, c=conn, p=puff:
				s.send_recpt(c, p))
		self.after_idle(lambda p=puff, c=conn: c.addpuff(p))

	def verify_update(self, conn, puff):
		self.after_idle(lambda p=puff, c=conn: c.puff_verified(p))
		
	# Send a return receipt if we're supposed to
	def send_recpt(self, conn, puff):
		if conn['send_receipts']:
			p = pygale.Puff()
			p.set_cat(puff.get('question/receipt'))
			p.set_text('answer/receipt', 'foo')
			p.set_text('notice/presence',
				self.presence_box.get(first=None))
			if Config.SENDER:
				p.set_text('message/sender', Config.SENDER)
			p.set_text('id/class', FUGU_VERSION_LONG)
			p.set_time('id/time', int(time.time()))
			p.set_text('id/instance', getinstance())
			# Sign return receipt
			p = p.sign_message(Config.SIGNER)
			# Encrypt if original message was signed
			old_signer = puff.get_signer(None)
			if old_signer is not None and old_signer not in \
				['*unsigned*', '*unverified*']:
				p = p.encrypt_message(old_signer,
					lambda p, s=self, orig=p: s.send_recpt2(p, orig))
			else:
				try:
					self._galeconn.transmit_puff(p)
				except pygale.PyGaleErr, e:
					# server connection died
					self.gale_error(e)
					# Assume retry is already in progress
					return
	
	def send_recpt2(self, puff, origpuff):
		if puff is None:
			self.gale_error('Unable to encrypt return receipt')
			puff = origpuff

		try:
			self._galeconn.transmit_puff(puff)
		except pygale.PyGaleErr, e:
			# server connection died
			# Assume retry is already in progress
			self.gale_error(e)

	# Send our key if it was requested
	def akd_request(self, reqpuff):
		reqkey = reqpuff.get('question/key', None)
		if reqkey is None or reqkey != Config.SIGNER:
			return

		me = pygale.export_pubkey(Config.SIGNER, lambda e, s=self:
			s.akd_request2(e))
	
	def akd_request2(self, export_key):
		if not export_key:
			return
		p = pygale.Puff()
		p.set_binary('answer/key', export_key)
		p.set_text('id/instance', getinstance())
		p.set_text('id/class', FUGU_VERSION_LONG)
		p.set_cat(pygale.id_category(Config.SIGNER, 'auth/key', ''))
		try:
			self._galeconn.transmit_puff(p)
		except pygale.PyGaleErr, e:
			# server connection died
			# Assume retry is already in progress
			self.gale_error(e)

	# Put some nice defaults into the puffwidget to reply to the currently
	# selected puff
	def reply(self):
		client = Config.SUBLIST[self._curscreen] 
		p = client.curpuff()
		if p is None: return

		REPLY_POPUP = 0
		if REPLY_POPUP:
			win = PuffWin(self)
		else:
			win = self._puffwidget

		# Figure out some nice defaults
		# Check for login/logout notices
		notice_re = re.compile('@([^/]*)/notice/([^/]*)/log')
		notice_groups = notice_re.match(p.get_cat())

		# See pygale/pygale.py:is_encrypted for comments
		# is_encrypted is obsolete as of 0.9.8c and only exists for
		# legacy puffs
		if (p.get_recipients() or p.is_encrypted()) and \
			p.get_signer(None) is not None:
			# If we're here, this was a private puff
			# Set destination to the set of all recipients plus the
			# signer, minus me
			dests = [p.get_signer()] + p.get_recipients()
			dests = filter(lambda x: x != Config.SIGNER, dests)
			if not dests:
				# If something bad happened and there is nothing left,
				# default to the signer
				dests = [p.get_signer()]
			else:
				hash_dests = {}
				for dest in dests:
					hash_dests[dest] = 1
				dests = hash_dests.keys()
			win.set_priv(string.join(dests, ' '))
			win.clear_pub()
		elif notice_groups:
			domain, user = notice_groups.groups()
			win.clear_pub()
			win.set_priv('%s@%s' % (user, domain))
		else:
			catlist = string.split(p.get_cat(), ':')
			catre = STRIPCATS
			catlist = filter(lambda x, r=catre: not re.match(r, x), catlist)
			if Config.JIHAD_FLAG:
				catlist = map(lambda x, s=self,
					delim=Config.JIHAD_DELIMITER:
					s._jihad_cat(x,delim), catlist)
			win.set_pub(string.join(catlist, ':'))
			win.clear_priv()
		win.set_subj(p.get('message/subject', ''))
		win.set_puff_focus()
	
	def _jihad_cat(self, cat, new_delim):
		# Jihad code contributed by <mshadow@ugcs.caltech.edu>
		#                  \1                             \2
		# matches either:  an initial / (condiment)       the condiment
		#                  a full directed 'special' cat  nothing
		#                  an initial directed prefix     the sub-cat
		#                  nothing                        the full category
		#  \1 is passed thru unchanged, \2 has [./] replaced with new_delim
		prefix = ''
		remain = cat
		# special: a regexp for directed sub-cats which shouldn't be changed
		special = 'user|auth|notice|server'
		my_re = re.compile('^(/|(?:@[^/]+/(?:(?:'+special+')/.*)?))?(.*)$')
		my_match = my_re.match(cat)
		if my_match:
			prefix = my_match.group(1)
			remain = my_match.group(2)
		if prefix is None: prefix = ''
		if remain is None: remain = ''
		remain = re.sub('[./]', new_delim, remain)
		return prefix+remain
		
	def reply_to_sender(self):
		client = Config.SUBLIST[self._curscreen] 
		p = client.curpuff()
		if p is None: return
		if p.get_signer(None) is None: return
		self._puffwidget.set_priv(p.get_signer())
		self._puffwidget.clear_pub()
		self._puffwidget.set_puff_focus()

	# Page up and down through the pufflist
	def key_up(self, *args):
		client = Config.SUBLIST[self._curscreen] 
		if not client.pufflist(): return
		index = client.curindex()
		if index is None:
			index = len(client) - 1
		if index > 0:
			index = index - 1
		client.set_curpuff(index)
		return 'break'

	def key_down(self, *args):
		client = Config.SUBLIST[self._curscreen] 
		if not client.pufflist(): return
		index = client.curindex()
		if index is None:
			index = len(client) - 1
		if index < len(client) - 1:
			index = index + 1
		client.set_curpuff(index)
		return 'break'

	def update_highlight(self):
		puff = Config.SUBLIST[self._curscreen].curpuff()
		if puff is None: return
		Config.SUBLIST[self._curscreen].widget().tag_config(
			'pufftag-%i' % hash(puff), background=Config.HIGHLIGHT)
	
	def update_autocomplete(self):
		self._puffwidget.update_autocomplete(Config.AUTOCOMPLETE)

	# Manipulate status bar text
	def set_status(self, text):
		self._status['text'] = text

	# Pop up the log message window
	def popup_logwin(self):
		self.logwin.deiconify()
		self.logwin.tkraise()

	# Pop up an about box
	def about(self):
		Pmw.aboutversion(FUGU_VERSION)
		Pmw.aboutcopyright('Copyright Tessa Lau 2000')
		Pmw.aboutcontact(
			'Tessa Lau\n' +
			'tlau@cs.washington.edu\n' +
			'http://ofb.net/fugu/')
		self.about = Pmw.AboutDialog(self, applicationname='Fugu')
		self.about.show()

	# Clear the puff fields
	def clearpuff(self):
		self._puffwidget.clear()
		self._puffwidget.set_pub_focus()
	
	# Tell the puffwidget to focus on the public field
	def set_pub_focus(self):
		self._puffwidget.set_pub_focus()

	# Detach this puff into its own window
	def detach(self):
		p = PuffWin(self)
		self.clearpuff()
	
	# Actions taken when send button is pressed
	def sendbutton(self, puffwidget=None, callback=None):
		if puffwidget is None:
			puffwidget = self._puffwidget

		heath_detector = re.compile('[^:]@')
		if heath_detector.search(puffwidget.pub()):
			d = tkMessageBox.askyesno('Heath alert',
				"Heath alert!  Your public category looks like "
				+ "a Gale ID.  Continue?")
			if not d: return
		if puffwidget.priv() and puffwidget.pub():
			d = tkMessageBox.askyesno('Crosspuff warning',
				"You're about to send a puff to a public category as" +
				" well as one or more private recipients.  Continue?")
			if not d: return
		# Check for leading whitespace in a category
		cats = puffwidget.pub()
		public_cats = string.split(cats, ':')
		badcats = filter(lambda x:
			x and ((x[0] not in (string.letters + '/@')) or
			(x[:4] == '/pub') or (x[:6] == '/local')),
			public_cats)
		if badcats:
			if len(badcats) == 1:
				s = 'The category %s has' % `badcats[0]`
			else:
				s = 'The categories %s have' % `string.join(badcats, ':')`
			s = s + ' funny characters at the beginning.  Continue?'
			d = tkMessageBox.askyesno('Questionable category', s)
			if not d: return
		if not puffwidget.puff():
			d = tkMessageBox.askyesno('Empty puff warning',
				"You're about to send an empty puff.  Continue?")
			if not d: return
		if not puffwidget.pub() and not puffwidget.priv():
			d = tkMessageBox.showerror(
				title='Error: no category',
				parent=self,
				message='No category has been specified.')
			return 1

		# Make the comboboxes save the categories in their histories
		puffwidget.save_state()
		if callback is None:
			callback = self.puff_sent
		ret = self.sendpuff(puffwidget.pub(),
			puffwidget.priv(), puffwidget.subj(), puffwidget.puff(),
			lambda r, cb=callback: cb(r))

	def puff_sent(self, result):
		if result:
			return
		self._puffwidget.clear_text()
		self._puffwidget.clear_priv()
		self._puffwidget.clear_subj()
		self._puffwidget.set_pub_focus()
	
	# Actions taken when option button is pressed
	def optbutton(self):
		w = OptionWin(self._puffwidget)

# ----------------------------------------------------------------------
# A combo widget containing pub/priv category entry widgets, and a text
# entry widget

class PuffWidget(Frame):
	def __init__(self, master=None, height=5, **kw):
		Frame.__init__(self, master)
		self._pubtext = kw.get('pubtext', '')
		self._privtext = kw.get('privtext', '')
		self._subjtext = kw.get('subjtext', '')
		self._pufftext = kw.get('pufftext', '')
		self._create_widgets(height)

	def _create_widgets(self, text_height):
		# Category entry
		self._pubcat = HistoryCombo(self,
			label_text='Public category:', labelpos='w',
			arrowbutton_takefocus=0, completion=Config.AUTOCOMPLETE)
		self._pubcat.component('entry').insert(0, self._pubtext)
		self._pubcat.component('entry').bind('<Shift-Key-Up>',
			lambda e, s=self, p=self._pubcat: s.drop_presence(p))
		self._pubcat.component('entry').bind('<Shift-Key-Down>',
			lambda e, s=self, p=self._pubcat: s.drop_presence(p))
		self._pubcat.component('entry').bind('<Tab>',
			lambda e, s=self._pubcat.component('entry'):
			s.selection_clear())
		self._pubcat.pack(side=TOP, fill=X, expand=0)

		self._privcat = HistoryCombo(self,
			label_text='People:', labelpos='w',
			arrowbutton_takefocus=0, completion=Config.AUTOCOMPLETE)
		self._privcat.component('entry').insert(0, self._privtext)
		self._privcat.component('entry').bind('<Shift-Key-Up>',
			lambda e, s=self, p=self._privcat: s.drop_presence(p))
		self._privcat.component('entry').bind('<Shift-Key-Down>',
			lambda e, s=self, p=self._privcat: s.drop_presence(p))
		self._privcat.component('entry').bind('<Tab>',
			lambda e, s=self._privcat.component('entry'):
			s.selection_clear())
		self._privcat.pack(side=TOP, fill=X, expand=0)

		self._subject = HistoryCombo(self, label_text='Subject:',
			labelpos='w', arrowbutton_takefocus=0,
			completion=Config.AUTOCOMPLETE)
		self._subject.component('entry').insert(0, self._subjtext)
		self._subject.component('entry').bind('<Shift-Key-Up>',
			lambda e, s=self, p=self._subject: s.drop_presence(p))
		self._subject.component('entry').bind('<Shift-Key-Down>',
			lambda e, s=self, p=self._subject: s.drop_presence(p))
		self._subject.component('entry').bind('<Tab>',
			lambda e, s=self._subject.component('entry'):
			s.selection_clear())
		if Config.SUBJECT_PROMPT:
			self._subject.pack(side=TOP, fill=X, expand=0)

		# Text of the puff
		self._puff = Pmw.ScrolledText(self, text_height=text_height,
			text_width=80, text_wrap='word', text_font=TEXT_FONT,
			vertscrollbar_takefocus=0)
		self._puff.pack(side=TOP, fill=BOTH, expand=1)
		self._puff.component('text').bind('<Tab>', self.tabForward)
		self._puff.insert('end', self._pufftext)
		
		if sys.platform != 'win32':
			# Don't bind these on windows; they're already bound
			self._bindkeys()

	def _bindkeys(self):
		# CUA-style bindings
		self.bind_generate('<Control-c>', '<<Copy>>', pub=0, priv=0)
		self.bind_generate('<Control-v>', '<<Paste>>', pub=0, priv=0)
	
	def update_autocomplete(self, newval):
		self._pubcat.configure(completion = newval)
		self._privcat.configure(completion = newval)
	
	def drop_presence(self, widget):
		widget.invoke()
		widget.component('listbox').focus_set()

	def ret_break(self, *args):
		return 'break'
	
	def save_state(self):
		self._pubcat._addHistory()
		self._privcat._addHistory()
		self._subject._addHistory()
	
	def bind(self, seq, func, add=None, pub=1, priv=1, text=1, subj=1):
		if pub:
			self._pubcat.component('entry').bind(seq, func, add)
		if priv:
			self._privcat.component('entry').bind(seq, func, add)
		if subj:
			self._subject.component('entry').bind(seq, func, add)
		if text:
			self._puff.component('text').bind(seq, func, add)
	
	def bind_generate(self, seq, event_str, pub=1, priv=1, text=1, subj=1):
		if pub:
			self._pubcat.component('entry').bind(seq,
				lambda e, w=self._pubcat.component('entry'), f=event_str:
				w.event_generate(f))
		if priv:
			self._privcat.component('entry').bind(seq,
				lambda e, w=self._privcat.component('entry'), f=event_str:
				w.event_generate(f))
		if subj:
			self._subject.component('entry').bind(seq,
				lambda e, w=self._privcat.component('entry'), f=event_str:
				w.event_generate(f))
		if text:
			self._puff.component('text').bind(seq,
				lambda e, w=self._puff.component('text'), f=event_str:
				w.event_generate(f))
	
	def show_subject(self):
		self._subject.pack(side=TOP, fill=X, expand=0, before=self._puff)
	
	def hide_subject(self):
		self._subject.forget()
	
	def call_and_break(self, func, event):
		func(event)
		return 'break'

	def tabForward(self, event):
		w = event.widget.tk_focusNext()
		w.focus_set()
		return 'break'

	def pub(self):
		return self._pubcat.get()
	
	def priv(self):
		return self._privcat.get()
	
	def subj(self):
		return self._subject.get()

	def text(self):
		return self._puff
	
	def puff(self):
		text = self._puff.get(1.0, 'end')
		# For some reason, the Text widget puts an extra \012 on the end
		# of this
		if text and text[-1] == '\012':
			text = text[:-1]
		if text and text[-1] == '\012':
			text = text[:-1]
		if text and text[-1] != '\012':
			text = text + '\012'
		return text
	
	def clear(self):
		self.clear_pub()
		self.clear_priv()
		self.clear_subj()
		self._puff.delete(1.0, 'end')
	
	def clear_text(self):
		self._puff.delete(1.0, 'end')
	
	def clear_pub(self):
		self._pubcat.component('entry').delete(0, 'end')

	def clear_priv(self):
		self._privcat.component('entry').delete(0, 'end')

	def clear_subj(self):
		self._subject.component('entry').delete(0, 'end')

	def set_pub(self, text):
		self.clear_pub()
		self._pubcat.component('entry').insert('end', text)
	
	def set_priv(self, text):
		self.clear_priv()
		self._privcat.component('entry').insert('end', text)

	def set_subj(self, text):
		self.clear_subj()
		self._subject.component('entry').insert('end', text)
	
	def set_puff_focus(self, *args):
		self._puff.component('text').focus_set()
	def set_pub_focus(self, *args):
		self._pubcat.component('entry').focus_set()
	def set_priv_focus(self, *args):
		self._privcat.component('entry').focus_set()
	def set_subj_focus(self, *args):
		self._subject.component('entry').focus_set()

# Create a toplevel clone of the master puffwidget
class PuffWin(Toplevel):
	def __init__(self, master):
		Toplevel.__init__(self)
		self._master = master
		# Make myself transient
		if hasattr(self, 'wm_transient'):
			self.wm_transient(master)
		elif hasattr(self, 'transient'):
			self.transient(master)
		# Buttons
		butframe = Frame(self)
		butframe.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		self._sendbutton = Button(butframe, text='Send (C-Enter)',
			command=self.sendbutton, takefocus=0)
		self._sendbutton.pack(side=LEFT, expand=0, fill=X)
		self._optbutton = Button(butframe, text='Options (C-O)',
			command=self.optbutton, takefocus=0, state='disabled')
		self._optbutton.pack(side=LEFT, expand=0, fill=X)
		self._abortbutton = Button(butframe, text='Abort (A-c)',
			command=self.destroy, takefocus=0)
		self._abortbutton.pack(side=LEFT, expand=0, fill=X)

		# Puff widget
		self._puffwidget = PuffWidget(self,
			pubtext=master._puffwidget.pub(),
			privtext = master._puffwidget.priv(),
			subjtext = master._puffwidget.subj(),
			pufftext = master._puffwidget.puff(), height=10, width=60)
		self._puffwidget.pack(side=BOTTOM, fill=BOTH, expand=1, padx=2,
			pady=2)

		# Hotkeys
		self._puffwidget.bind('<Control-Return>', self.control_return)
		self.bind('<Alt-c>', lambda e, s=self:
			s._abortbutton.invoke())
		self._puffwidget.set_puff_focus()
	
	# This is more than just invoking the send button; it also has to
	# return 'break' so the return doesn't get processed and add a newline
	# to the message.
	def control_return(self, event):
		self.sendbutton()
		return 'break'

	def sendbutton(self, *args):
		self._master.sendbutton(self._puffwidget, self.puff_sent)

	def puff_sent(self, result):
		if result:
			return
		self.destroy()
	
	def optbutton(self):
		w = OptionWin(self._puffwidget)
	
	def set_pub(self, *args):
		return apply(self._puffwidget.set_pub, args)
	def set_priv(self, *args):
		return apply(self._puffwidget.set_priv, args)
	def set_subj(self, *args):
		return apply(self._puffwidget.set_subj, args)
	def clear_pub(self, *args):
		return apply(self._puffwidget.clear_pub, args)
	def clear_priv(self, *args):
		return apply(self._puffwidget.clear_priv, args)
	def clear_subj(self, *args):
		return apply(self._puffwidget.clear_subj, args)
	def set_puff_focus(self, *args):
		return apply(self._puffwidget.set_puff_focus, args)
	def set_pub_focus(self, *args):
		return apply(self._puffwidget.set_pub_focus, args)
	def set_priv_focus(self, *args):
		return apply(self._puffwidget.set_priv_focus, args)
	def set_subj_focus(self, *args):
		return apply(self._puffwidget.set_subj_focus, args)

class OptionWin(Toplevel):
	def __init__(self, puffwidget):
		Toplevel.__init__(self)
		self._puff = puffwidget
	
		# Widgets
		butframe = Frame(self)
		butframe.pack(side=BOTTOM, expand=1, fill=X)

		f = Frame(self)
		f.pack(side=LEFT, expand=1, fill=BOTH)
		self.encrypt = IntVar()
		self.encrypt.trace('w', self.enc_cb)
		r1 = Radiobutton(f, text='Encrypt to recipients',
			variable=self.encrypt, value=0, anchor='w')
		r1.pack(side=TOP, fill=X)
		r2 = Radiobutton(f, text='Do not encrypt', value=1,
			variable=self.encrypt, anchor='w')
		r2.pack(side=TOP, fill=X)
		r3 = HistoryCombo(f,
			label_text='Encrypt to:', labelpos='w',
			completion=Config.AUTOCOMPLETE)
		r3.pack(side=TOP, fill=X)
		
		but1 = Button(butframe, text='OK', command=self.ok)
		but1.pack(side=LEFT, expand=1, fill=X)
		but2 = Button(butframe, text='Cancel', command=self.destroy)
		but2.pack(side=LEFT, expand=1, fill=X)
		
		self.grab_set()
		self.wait_window(self)
	
	def enc_cb(self, *args):
		if self.encrypt.get() == 2:
			self.entry['state'] = 'normal'
		else:
			self.entry['state'] = 'disabled'
	
	# TODO: add these options to the puff
	def ok(self):
		self.encrypt.get()
		self.destroy()
	
class ConfigPop(Toplevel):
	def __init__(self, master):
		Toplevel.__init__(self)
		if hasattr(self, 'wm_transient'):
			self.wm_transient(master)
		elif hasattr(self, 'transient'):
			self.transient(master)
		self._master = master

		# OK/Cancel buttons
		butframe = Frame(self)
		butframe.pack(side='bottom', expand=1, fill=X)
		ok = Button(butframe, text="OK", command=self.ok)
		ok.pack(side=LEFT, expand=1, fill=X)
		cancel = Button(butframe, text="Cancel", command=self.destroy)
		cancel.pack(side=LEFT, expand=1, fill=X)
	
		f = Frame(self)
		f.pack(side='left', expand=1, fill=BOTH)
		g = Frame(self)
		g.pack(side='right', expand=1, fill=BOTH)

		# Puff sender
		self.sender = Pmw.EntryField(f, labelpos='w',
			value=Config.SENDER, label_text = 'Puff sender:')
		self.sender.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Signer
		self.signer = Pmw.EntryField(f, labelpos='w',
			value=Config.SIGNER, label_text = 'Puff signer:')
		self.signer.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Puff history
		self.scrollback = Pmw.EntryField(f,
			labelpos='w', value=Config.SCROLLBACK,
			label_text = 'Puff history:',
			validate={'validator':'integer', 'min':1, 'max':1000})
		self.scrollback.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Puff persist interval
		self.persist = Pmw.EntryField(f,
			labelpos='w', value=Config.PERSIST_INTERVAL,
			label_text = 'Save puff history (secs)',
			validate={'validator':'integer', 'min':0})
		self.persist.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Cc self on private puffs
		self.cc_var = BooleanVar()
		self.cc_var.set(Config.CC_SELF)
		cc = Checkbutton(f, text='Send copy of private puffs to self',
			variable=self.cc_var, anchor='w')
		cc.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Presence status
		self.presence_var = BooleanVar()
		self.presence_var.set(Config.BROADCAST_PRESENCE)
		self.presence = Checkbutton(f, text='Broadcast presence status',
			variable=self.presence_var, onval=1, offval=0, anchor='w')
		self.presence.pack(fill=X, expand=1, padx=2, pady=2)

		# Save width on exit
		self.save_width_var = BooleanVar()
		self.save_width_var.set(Config.SAVE_PANE_WIDTH)
		button = Checkbutton(f, text='Save window width on exit',
			variable=self.save_width_var, onval=1, offval=0, anchor='w')
		button.pack(fill=BOTH, expand=1, padx=2, pady=2)

		# Map/unmap puffbuttonbar
		self.map_var = BooleanVar()
		self.map_var.set(Config.MAP_PUFFBUTTONS)
		self.map_var.trace('w', self.toggle_mapbuttons)
		button = Checkbutton(f, text='Display puff button bar',
			variable=self.map_var, onval=1, offval=0, anchor='w')
		button.pack(fill=BOTH, expand=1, padx=2, pady=2)

		# Map/unmap subject prompt
		self.subj_var = BooleanVar()
		self.subj_var.set(Config.SUBJECT_PROMPT)
		self.subj_var.trace('w', self.toggle_subject)
		button = Checkbutton(f, text='Prompt for puff subject',
			variable=self.subj_var, onval=1, offval=0, anchor='w')
		button.pack(fill=BOTH, expand=1, padx=2, pady=2)

		# Whether to display thumbnails
		self.thumb_var = BooleanVar()
		self.thumb_var.set(Config.SHOW_THUMBS)
		button = Checkbutton(f, text='Display thumbnail images',
			variable = self.thumb_var, onval=1, offval=0, anchor='w')
		button.pack(fill=BOTH, expand=1, padx=2, pady=2)

		# Autocomplete category strings
		self.autocomplete_var = BooleanVar()
		self.autocomplete_var.set(Config.AUTOCOMPLETE)
		button = Checkbutton(f, text='Autocomplete category strings',
			variable=self.autocomplete_var, onval=1, offval=0,
			anchor='w')
		button.pack(fill=BOTH, expand=1, padx=2, pady=2)

		# URL command
		group = Pmw.Group(g, tag_text='URL command')
		group.pack(side=TOP, fill=BOTH, expand=1, padx=2, pady=2)
		self.urlcommand = Entry(group.interior(), width=30)
		self.urlcommand.insert(0, Config.URLLOAD)
		self.urlcommand.pack(side=LEFT, expand=1, fill=X, padx=2, pady=2)

		# Thumbnail URL
		group = Pmw.Group(g, tag_text='URL to my thumbnail')
		group.pack(side=TOP, fill=BOTH, expand=1, padx=2, pady=2)
		self.thumburl = Entry(group.interior(), width=30)
		self.thumburl.insert(0, Config.THUMBNAIL_URL)
		self.thumburl.pack(side=LEFT, expand=1, fill=X, padx=2, pady=2)
 
 		# JTR's Jihad patch
		self.jihad_flag = BooleanVar()
		self.jihad_flag.set(Config.JIHAD_FLAG)
		group = Pmw.Group(g, tag_text='Convert intracategory delimiters',
			tag_pyclass = Checkbutton,
			tag_variable = self.jihad_flag, tag_onval=1, tag_offval=0)
		group.pack(fill=BOTH, expand=1, padx=2, pady=2)

		DELIMITER_CHOICES = ['.', '/']
		self.jihad_delimiter = Pmw.OptionMenu(group.interior(),
			label_text='Switch separators to:',
			labelpos = 'w',
			items=DELIMITER_CHOICES,
			initialitem=Config.JIHAD_DELIMITER)
		self.jihad_delimiter.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Highlighted puff color
		group = Pmw.Group(g, tag_text='Puff highlight color')
		group.pack(side=TOP, fill=BOTH, expand=1, padx=2, pady=2)
		self.newcolor = Config.HIGHLIGHT
		self.colorbutton = Button(group.interior(), text='Choose color',
			background=Config.HIGHLIGHT, command=self.getcolor)
		self.colorbutton.pack(expand=0, fill=X)

	def getcolor(self):
		color = tkColorChooser.askcolor(initialcolor=self.newcolor)[1]
		if color is not None:
			self.newcolor = color
			self.colorbutton['background'] = self.newcolor

	def ok(self, *args):
		# TODO: check for valid scrollback value
		Config.BROADCAST_PRESENCE = self.presence_var.get()
		Config.SENDER = self.sender.get()
		Config.SIGNER = self.signer.get()
		Config.CC_SELF = self.cc_var.get()
		Config.SCROLLBACK = string.atoi(self.scrollback.get())
		Config.HIGHLIGHT = self.newcolor
		Config.SAVE_PANE_WIDTH = self.save_width_var.get()
		Config.URLLOAD = self.urlcommand.get()
		Config.MAP_PUFFBUTTONS = self.map_var.get()
		Config.SUBJECT_PROMPT = self.subj_var.get()
		Config.JIHAD_FLAG = self.jihad_flag.get()
		if Pmw.version() > '0.8.1':
			Config.JIHAD_DELIMITER = self.jihad_delimiter.getcurselection()
		else:
			Config.JIHAD_DELIMITER = self.jihad_delimiter.get()
		Config.SHOW_THUMBS = self.thumb_var.get()
		Config.THUMBNAIL_URL = self.thumburl.get()
		Config.AUTOCOMPLETE = self.autocomplete_var.get()

		# Updates caused by these changes
		self._master.update_highlight()
		self._master.update_autocomplete()

		# Save configuration and go away
		Config.save(FUGUDIR)
		self.destroy()
	
	def toggle_mapbuttons(self, *args):
		if self.map_var.get():
			self._master.puffbuts.pack(side=TOP, fill=X, expand=0,
				padx=2, before=self._master._puffwidget)
		else:
			self._master.puffbuts.forget()
	
	def toggle_subject(self, *args):
		if self.subj_var.get():
			self._master.show_subject()
		else:
			self._master.hide_subject()

class LogWin(Toplevel):
	def __init__(self, master=None):
		Toplevel.__init__(self)
		self._master = master
		self.protocol('WM_DELETE_WINDOW', self.withdraw)

		l = Label(self, text='Gale notices and warnings:', anchor='w')
		l.pack(side=TOP, expand=0, fill=X)

		self.text = Pmw.ScrolledText(self)
		self.text.pack(side=TOP, expand=1, fill=BOTH)
		self.text.configure(text_state='disabled')

		c = Button(self, text='Close', command=self.withdraw)
		c.pack(side=TOP, expand=0, fill=X)

	def add_message(self, msg):
		self.text.configure(text_state='normal')
		self.text.insert('end', msg + '\n')
		# delete old messages
		while int(float(self.text.index('end'))) >= LOG_MESG_BUFFER + 3:
			self.text.delete('1.0', '2.0')
		self.text.see('end')
		self.text.configure(text_state='disabled')

class ScreenPop(Toplevel):
	def __init__(self, master=None):
		Toplevel.__init__(self)
		if master:
			if hasattr(self, 'wm_transient'):
				self.wm_transient(master)
			elif hasattr(self, 'transient'):
				self.transient(master)
		self.protocol('WM_DELETE_WINDOW', self.withdraw)
		self._master = master
		self.oldsel = None

		close = Button(self, text='Close', command=self.close)
		close.pack(side=BOTTOM, expand=0, fill=X, padx=2, pady=2)

		f = Frame(self)
		f.pack(side=LEFT, expand=1, fill=BOTH)
		label = Label(f, text='Screens:', anchor='w')
		label.pack(side=TOP, fill=X, expand=0, padx=2, pady=2)
		listbox = Pmw.ScrolledListBox(f, selectioncommand=self.select,
			listbox_selectmode='single', listbox_takefocus=0,
			listbox_exportselection=0)
		listbox.pack(side=LEFT, fill=BOTH, expand=1, padx=2, pady=2)
		self.listbox = listbox

		for c in Config.SUBLIST:
			listbox.insert('end', c['name'])
		
		# Manipulation of screen list
		g = Frame(f)
		g.pack(side=LEFT, expand=0, fill=BOTH)
		addbut = Button(g, text='New screen', command=self.new)
		addbut.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		cpbut = Button(g, text='Copy screen', command=self.copy)
		cpbut.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		delbut = Button(g, text='Delete screen', command=self.delete)
		delbut.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		upbut = Button(g, text='Move up', command=self.up)
		upbut.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		downbut = Button(g, text='Move down', command=self.down)
		downbut.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# Options per screen
		# Name and subscription
		h = Frame(self, relief='groove', borderwidth=2)
		h.pack(side=TOP, expand=1, fill=BOTH, padx=2, pady=2)
		i = Frame(h)
		i.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		label = Label(i, text='Name:', anchor='w')
		label.pack(side=LEFT, expand=0, fill=X)
		self.namevar = StringVar()
		self.name_change_cb = self.namevar.trace('w', self.name_change)
		name = Entry(i, textvariable=self.namevar)
		name.pack(side=LEFT, expand=0, fill=X)
		self.dosubvar = BooleanVar()
		group = Pmw.Group(h, tag_pyclass = Checkbutton,
			tag_variable=self.dosubvar,
			tag_text = 'Subscribe to:')
		group.pack(side=TOP, expand=1, fill=BOTH, padx=2, pady=2)
		self.sub = Pmw.ScrolledText(group.interior(), text_width=40,
			text_height=5, text_wrap='word', text_font=TEXT_FONT)
		self.sub.pack(side=TOP, expand=1, fill=BOTH, padx=2, pady=2)
		# Whether to save puff history
		self.savevar = BooleanVar()
		save = Checkbutton(h, text='Save puff history on exit',
			variable=self.savevar, anchor='w')
		save.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		# Whether to send return receipts
		self.recptvar = BooleanVar()
		recpt = Checkbutton(h, text='Send return receipts to private puffs',
			variable=self.recptvar, anchor='w')
		recpt.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		# Log sent puffs
		LOG_CHOICES = ['Nothing', 'Log private to screen',
			'Log all to screen']
		self.log = Pmw.OptionMenu(h,
			label_text='Sent puff handling:',
			labelpos = 'w',
			items=LOG_CHOICES)
		self.log.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)

		# On new puff
		group = Pmw.Group(h, tag_text = 'On new puff:')
		group.pack(side=TOP, expand=0, fill=X, padx=2, pady=2)
		f = group.interior()
		f.columnconfigure(0, weight=1)
		f.columnconfigure(1, weight=1)
		f.rowconfigure(0, weight=1)
		f.rowconfigure(1, weight=1)
		f.rowconfigure(2, weight=1)

		self.beepvar = BooleanVar()
		beep = Checkbutton(group.interior(), text='Beep',
			variable=self.beepvar, anchor='w')
		beep.grid(row=0, column=0, padx=2, pady=2, sticky='we')

		self.highlightvar = BooleanVar()
		highlight = Checkbutton(group.interior(), text='Highlight button',
			variable=self.highlightvar, anchor='w')
		highlight.grid(row=0, column=1, padx=2, pady=2, sticky='we')

		self.liftvar = BooleanVar()
		lift = Checkbutton(group.interior(), text='Lift screen to top',
			variable=self.liftvar, anchor='w')
		lift.grid(row=1, column=0, padx=2, pady=2, sticky='we')

		self.deiconvar = BooleanVar()
		deicon = Checkbutton(group.interior(), text='Deiconify window',
			variable=self.deiconvar, anchor='w')
		deicon.grid(row=1, column=1, padx=2, pady=2, sticky='we')

		self.cmdvar = BooleanVar()
		cmdgrp = Pmw.Group(group.interior(), tag_text = 'Run command:',
			tag_pyclass=Checkbutton, tag_variable=self.cmdvar,
			tag_command=self.toggle_cmd)
		cmdgrp.grid(row=2, column=0, columnspan=2,
			padx=2, pady=2, sticky='we')
		self.cmd = Entry(cmdgrp.interior())
		self.cmd.pack(fill=X, padx=2, pady=2)
		self.cmd['state'] = 'disabled'

		# Initialize listbox selection
		self.listbox.selection_set(0)
		self.select()
	
	def set_screen(self, index):
		self.listbox.component('listbox').selection_clear(0, 'end')
		self.listbox.selection_set(index)
		self.select()
	
	def name_change(self, name1, name2, op):
		index = self.get_sel_as_index()
		conf = self.get_sel_as_conf()
		name = self.namevar.get()
		self.listbox.delete(index)
		self.listbox.insert(index, name)
		self.listbox.component('listbox').selection_clear(0, 'end')
		self.listbox.selection_set(index)
		conf['name'] = name
		self._master.change_button_name(index, name)
		self._master.regen_screenmenu()
	
	def toggle_cmd(self):
		if self.cmdvar.get():
			self.cmd['state'] = 'normal'
		else:
			self.cmd['state'] = 'disabled'

	def up(self):
		index = self.get_sel_as_index()
		if index is None: return
		if index == 0:
			self.bell()
			return
		sel = Config.SUBLIST[index]
		self.save_current_values(sel)
		self.listbox.delete(index)
		self.listbox.insert(index-1, sel['name'])
		self.listbox.component('listbox').selection_clear(0, 'end')
		self.listbox.selection_set(index-1)
		Config.SUBLIST[index:index+1] = []
		Config.SUBLIST.insert(index-1, sel)
		self.select()
		self._master.move_button_up(index)
		self._master.regen_screenmenu()
	
	def down(self):
		index = self.get_sel_as_index()
		if index is None: return
		if index == len(self.listbox.get(0, 'end'))-1:
			self.bell()
			return
		sel = Config.SUBLIST[index]
		self.save_current_values(sel)
		self.listbox.delete(index)
		self.listbox.insert(index+1, sel['name'])
		self.listbox.component('listbox').selection_clear(0, 'end')
		self.listbox.selection_set(index+1)
		Config.SUBLIST[index:index+1] = []
		Config.SUBLIST.insert(index+1, sel)
		self.select()
		self._master.move_button_down(index)
		self._master.regen_screenmenu()

	def new(self):
		new = connection.Connection('New screen', 'pub')
		Config.SUBLIST.append(new)
		text = self._master.make_text_widget()
		new.setwidget(self._master, text)
		new.connect()
		text.insert(1.0, new['sub'] + '\n')
		self._master.add_button(new)
		self._master.regen_screenmenu()
		self.listbox.insert('end', new['name'])
		self.listbox.component('listbox').selection_clear(0, 'end')
		self.listbox.selection_set('end')
		self.select()
	
	def copy(self):
		sel = self.get_sel_as_conf()
		if sel is None: return
		new = connection.Connection(sel['name'] + ' copy', sel['sub'])
		Config.SUBLIST.append(new)
		text = self._master.make_text_widget()
		text.insert(1.0, new['sub'] + '\n')
		new.setwidget(self._master, text)
		new.connect()
		self._master.add_button(new)
		self._master.regen_screenmenu()
		self.listbox.insert('end', new['name'])
		self.listbox.component('listbox').selection_clear(0, 'end')
		self.listbox.selection_set('end')
		self.select()
	
	def delete(self):
		index = self.get_sel_as_index()
		delwindow = Config.SUBLIST[index]
		if len(Config.SUBLIST) == 1:
			m = tkMessageBox.showerror(
				title="Error: can't delete last screen",
				parent=self,
				message='At least one screen must be defined.')
			return

		# Upcall into parent if this is currently displayed
		if self._master.is_curscreen(delwindow):
			if index == 0:
				switchindex = 1
			else:
				switchindex = index - 1
			self._master.switch_screen(switchindex)

		Config.SUBLIST[index].disconnect()
		del Config.SUBLIST[index]
		self.listbox.delete(index)
		self.listbox.component('listbox').selection_clear(0, 'end')
		self._master.del_button(index)
		self._master.regen_screenmenu()
		if index >= len(self.listbox.get(0, 'end')):
			index = len(self.listbox.get(0, 'end')) - 1
		self.listbox.selection_set(index)
		self.select()
	
	def get_sel_as_index(self):
		selection = map(string.atoi, self.listbox.curselection())
		if len(selection) == 0: return None
		return selection[0]

	def get_sel_as_conf(self):
		selindex = self.get_sel_as_index()
		if selindex is None: return None
		return Config.SUBLIST[selindex]

	def save_current_values(self, conf):
		# Save old values
		reconnect = 0
		if conf['sub'] != self.sub.get(1.0, 'end-1c') or\
			conf['dosub'] != self.dosubvar.get():
			reconnect = 1
		conf['name'] = self.namevar.get()
		conf['dosub'] = self.dosubvar.get()
		conf['sub'] = self.sub.get(1.0, 'end-1c')
		conf['beep'] = self.beepvar.get()
		conf['highlight'] = self.highlightvar.get()
		conf['lift'] = self.liftvar.get()
		conf['deiconify'] = self.deiconvar.get()
		conf['savepuffs'] = self.savevar.get()
		conf['send_receipts'] = self.recptvar.get()
		if Pmw.version() > '0.8.1':
			conf['logsent'] = \
				self.log.index(self.log.getcurselection())
		else:
			conf['logsent'] = self.log.index(self.log.get())
		if self.cmdvar.get():
			conf['command'] = self.cmd.get()
		else:
			conf['command'] = None
		if reconnect:
			conf.reconnect()

	def select(self):
		sel = self.get_sel_as_conf()
		if sel is None: return
		if self.oldsel is not None and self.oldsel != sel:
			self.save_current_values(self.oldsel)
		self.namevar.trace_vdelete('w', self.name_change_cb)
		self.namevar.set(sel['name'])
		self.name_change_cb = self.namevar.trace('w', self.name_change)
		self.sub.delete(1.0, 'end')
		self.sub.insert('end', sel['sub'])
		self.dosubvar.set(sel['dosub'])
		self.beepvar.set(sel['beep'])
		self.highlightvar.set(sel['highlight'])
		self.liftvar.set(sel['lift'])
		self.deiconvar.set(sel['deiconify'])
		self.savevar.set(sel['savepuffs'])
		self.recptvar.set(sel['send_receipts'])
		self.log.invoke(sel['logsent'])
		self.cmd['state'] = 'normal'
		self.cmd.delete(0, 'end')
		if sel['command'] is None:
			self.cmdvar.set(0)
			self.cmd['state'] = 'disabled'
		else:
			self.cmdvar.set(1)
			self.cmd.insert('end', sel['command'])
		self.oldsel = sel
	
	def close(self):
		sel = self.get_sel_as_conf()
		if sel is not None:
			self.save_current_values(sel)
		if len(Config.SUBLIST) == 0:
			m = tkMessageBox.showerror(
				title='Error: no subscriptions',
				parent=self,
				message='At least one screen must be defined.')
			return

		name_collisions = {}
		for conn in Config.SUBLIST:
			if name_collisions.has_key(conn['name']):
				m = tkMessageBox.showerror(
					title='Error: duplicate screen names',
					parent=self,
					message='No two screens may have the same name.')
				return
			name_collisions[conn['name']] = 1
		
		# Save configuration and go away
		Config.save(FUGUDIR)
		self.withdraw()

def main():
	pygale.init()
	sys.exitfunc = pygale.shutdown

	# Command line options
	opts, args = getopt.getopt(sys.argv[1:], 'd')
	for (opt, val) in opts:
		if opt == '-d':
			global DEBUG
			DEBUG = 1
		else:
			print 'Unsupported argument:', opt

	# Fugu directory
	global FUGUDIR
	FUGUDIR = pygale.gale_env.get('FUGU_DIR')
	thumbcache.FUGUDIR = FUGUDIR
	global Config
	Config = config.ConfigClass()
	# Load configuration
	Config.load(FUGUDIR)
	Config.load_pufflog(FUGUDIR)
	# Propagate configuration everywhere it has to go
	thumbcache.Config = Config
	connection.Config = Config
	root = Tk(None, None, "Fugu")
	Pmw.initialise(root)
	if DEBUG and sys.version >= '2.0':
		import gc
		print 'Setting debugging GC'
		gc.set_debug(gc.DEBUG_LEAK)
	m = GaleWin(root)
	m.mainloop()

if __name__ == '__main__':
	main()

#vim:ts=5:sw=5:sts=5:
